"""Deep learning layers"""
from typing import Optional, Sequence

import tensorflow as tf
import tensorflow.keras.backend as kb
from tensorflow.keras import activations, constraints, initializers, regularizers
from tensorflow.keras.layers import Layer


class WeightedAverageLayer(Layer):
    r"""
    Weight average the features using weights

    result= \sum{w_i^a * value_i} / \sum{w_i^a}

    """

    def __init__(self, alpha: float = 1.0, **kwargs):
        """
        Args:
            alpha (float): exponent in weighting
        """
        super().__init__(**kwargs)
        self.alpha = alpha

    def build(self, input_shape: Sequence) -> None:
        """
        Build the layer

        Args:
            input_shape (tuple): input shape tuple
        """
        self.built = True

    def call(self, inputs: Sequence, mask: Optional[tf.Tensor] = None) -> tf.Tensor:
        """
        Core logic of the layer

        Args:
            inputs (tuple): input tuple of length 3
            mask (tf.Tensor): not used here
        """
        # prop: [1, n, n_feature]
        # weight: [1, n]
        # index： [1, n]
        prop, weights, index = inputs
        expo_weights = weights**self.alpha
        prop = prop * expo_weights[:, :, None]
        prop_sum = self.reduce_sum(prop, index, perm=[1, 0, 2])
        weight_sum = self.reduce_sum(expo_weights, index, perm=[1, 0])
        return prop_sum / weight_sum[:, :, None]

    @staticmethod
    def reduce_sum(prop: tf.Tensor, index: tf.Tensor, perm: Sequence) -> tf.Tensor:
        """
        Reduce sum the tensors using index

        Args:
            prop (tf.Tensor): tensor with shape [1, n, ...]
            index (tf.Tensor): integer tensor with shape [1, n]
            perm (list): permutation for transpose.
        """
        index = tf.reshape(index, (-1,))
        prop = tf.transpose(a=prop, perm=perm)
        out = tf.math.segment_sum(prop, index)
        out = tf.transpose(a=out, perm=perm)
        return out

    @staticmethod
    def compute_output_shape(input_shape: Sequence) -> tuple:
        """
        Compute output shape from input shape
        Args:
            input_shape (tuple/list): input shape list
        """
        prop_shape = input_shape[0]
        return prop_shape[0], None, prop_shape[-1]

    def get_config(self) -> dict:
        """
        Get layer configuration
        """
        config = {"alpha": self.alpha}
        base_config = super().get_config()
        config.update(base_config)
        return config  # type: ignore


class WeightedSet2Set(Layer):
    """
    This is a modified version from megnet.layers.readout.Set2Set.
    Here, in addition to taking features and indices as inputs, we also
    take a weight tensor. The input to the core logic is
    [features, weights, indices]
    """

    def __init__(
        self,
        T=3,
        n_hidden=512,
        activation=None,
        activation_lstm="tanh",
        recurrent_activation="hard_sigmoid",
        kernel_initializer="glorot_uniform",
        recurrent_initializer="orthogonal",
        bias_initializer="zeros",
        use_bias=True,
        unit_forget_bias=True,
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        **kwargs,
    ):
        """
        Args:
            T: (int) recurrent step
            n_hidden: (int) number of hidden units
            activation: (str or object) activation function
            activation_lstm: (str or object) activation function for lstm
            recurrent_activation: (str or object) activation function for recurrent step
            kernel_initializer: (str or object) initializer for kernel weights
            recurrent_initializer: (str or object) initializer for recurrent weights
            bias_initializer: (str or object) initializer for biases
            use_bias: (bool) whether to use biases
            unit_forget_bias: (bool) whether to use basis in forget gate
            kernel_regularizer: (str or object) regularizer for kernel weights
            recurrent_regularizer: (str or object) regularizer for recurrent weights
            bias_regularizer: (str or object) regularizer for biases
            kernel_constraint: (str or object) constraint for kernel weights
            recurrent_constraint: (str or object) constraint for recurrent weights
            bias_constraint:(str or object) constraint for biases
            kwargs: other inputs for keras Layer class
        """
        super().__init__(**kwargs)
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

        self.activation_lstm = activations.get(activation_lstm)
        self.recurrent_activation = activations.get(recurrent_activation)
        self.recurrent_initializer = initializers.get(recurrent_initializer)
        self.unit_forget_bias = unit_forget_bias
        self.recurrent_regularizer = regularizers.get(recurrent_regularizer)
        self.recurrent_constraint = constraints.get(recurrent_constraint)
        self.T = T
        self.n_hidden = n_hidden

    def build(self, input_shape: Sequence) -> None:
        """
        Build the output shape from input shapes
        Args:
            input_shape (tuple/list): input shape list
        """
        feature_shape, weight_shape, index_shape = input_shape
        self.m_weight = self.add_weight(
            shape=(feature_shape[-1], self.n_hidden),
            initializer=self.kernel_initializer,
            name="x_to_m_weight",
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )
        if self.use_bias:
            self.m_bias = self.add_weight(
                shape=(self.n_hidden,),
                initializer=self.bias_initializer,
                name="x_to_m_bias",
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        else:
            self.m_bias = None

        self.recurrent_kernel = self.add_weight(
            shape=(2 * self.n_hidden, 4 * self.n_hidden),
            name="recurrent_kernel",
            initializer=self.recurrent_initializer,
            regularizer=self.recurrent_regularizer,
            constraint=self.recurrent_constraint,
        )
        if self.use_bias:
            if self.unit_forget_bias:

                def bias_initializer(_, *args, **kwargs):
                    return kb.concatenate(
                        [
                            self.bias_initializer((self.n_hidden,), *args, **kwargs),
                            initializers.Ones()((self.n_hidden,), *args, **kwargs),
                            self.bias_initializer((self.n_hidden * 2,), *args, **kwargs),
                        ]
                    )

            else:
                bias_initializer = self.bias_initializer
            self.recurrent_bias = self.add_weight(
                shape=(self.n_hidden * 4,),
                name="recurrent_bias",
                initializer=bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        else:
            self.recurrent_bias = None
        self.built = True

    def compute_output_shape(self, input_shape):
        """
        Compute output shapes from input shapes
        Args:
            input_shape (sequence of tuple): input shapes

        Returns: sequence of tuples output shapes

        """
        feature_shape, index_shape = input_shape
        return feature_shape[0], None, 2 * self.n_hidden

    def call(self, inputs, mask=None):
        """
        Core logic of the layer

        Args:
            inputs (tuple): input tuple of length 3
            mask (tf.Tensor): not used here
        """
        features, weights, feature_graph_index = inputs
        feature_graph_index = tf.reshape(feature_graph_index, (-1,))
        counts = tf.math.bincount(feature_graph_index)
        n_count = tf.shape(counts)[0]
        n_feature = tf.shape(features)[0]
        m = kb.dot(features, self.m_weight)
        if self.use_bias:
            m += self.m_bias

        self.h = tf.zeros(tf.stack([n_feature, n_count, self.n_hidden]))
        self.c = tf.zeros(tf.stack([n_feature, n_count, self.n_hidden]))
        q_star = tf.zeros(tf.stack([n_feature, n_count, 2 * self.n_hidden]))
        for i in range(self.T):
            self.h, c = self._lstm(q_star, self.c)
            e_i_t = tf.reduce_sum(input_tensor=m * tf.repeat(self.h, repeats=counts, axis=1), axis=-1)
            maxes = tf.math.segment_max(e_i_t[0], feature_graph_index)
            maxes = tf.repeat(maxes, repeats=counts)
            e_i_t -= tf.expand_dims(maxes, axis=0)
            exp = tf.exp(e_i_t) * weights
            seg_sum = tf.transpose(
                a=tf.math.segment_sum(tf.transpose(a=exp, perm=[1, 0]), feature_graph_index), perm=[1, 0]
            )
            seg_sum = tf.expand_dims(seg_sum, axis=-1)
            interm = tf.repeat(seg_sum, repeats=counts, axis=1)
            a_i_t = exp / interm[..., 0]
            r_t = tf.transpose(
                a=tf.math.segment_sum(
                    tf.transpose(a=tf.multiply(m, a_i_t[:, :, None]), perm=[1, 0, 2]), feature_graph_index
                ),
                perm=[1, 0, 2],
            )
            q_star = kb.concatenate([self.h, r_t], axis=-1)
        return q_star

    def _lstm(self, h, c):
        # lstm implementation here
        z = kb.dot(h, self.recurrent_kernel)
        if self.use_bias:
            z += self.recurrent_bias
        z0 = z[:, :, : self.n_hidden]
        z1 = z[:, :, self.n_hidden : 2 * self.n_hidden]
        z2 = z[:, :, 2 * self.n_hidden : 3 * self.n_hidden]
        z3 = z[:, :, 3 * self.n_hidden :]
        i = self.recurrent_activation(z0)
        f = self.recurrent_activation(z1)
        # print(z.shape, f.shape, c.shape, z2.shape)
        c = f * c + i * self.activation_lstm(z2)
        o = self.recurrent_activation(z3)
        h = o * self.activation_lstm(c)
        return h, c

    def get_config(self):
        """
         Part of keras layer interface, where the signature is converted into a dict
        Returns:
            configurational dictionary
        """
        config = {
            "T": self.T,
            "n_hidden": self.n_hidden,
            "activation": activations.serialize(self.activation),
            "activation_lstm": activations.serialize(self.activation_lstm),
            "recurrent_activation": activations.serialize(self.recurrent_activation),
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "recurrent_initializer": initializers.serialize(self.recurrent_initializer),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "use_bias": self.use_bias,
            "unit_forget_bias": self.unit_forget_bias,
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "recurrent_regularizer": regularizers.serialize(self.recurrent_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "recurrent_constraint": constraints.serialize(self.recurrent_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint),
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""
Simple numba utility.
Some functions can excelerated substantially with numba
"""
from typing import Callable

try:
    from numba import njit  # type: ignore
except ModuleNotFoundError:

    def njit(func: Callable) -> Callable:
        """
        Dummy decorator, returns the original function
        Args:
            func (Callable): function to be wrapped

        Returns: decorated function

        """
        return func



"""
Data split
"""
from typing import Optional, Tuple

import numpy as np
from monty.json import MSONable
from sklearn.utils import shuffle


class DataSplitter(MSONable):
    """
    Data splitter base class
    """

    def split(self, mat_ids, **kwargs):
        """
        Split the mat_ids, optionally one can provide
        targets. This is useful in stratified split.

        Args:
            mat_ids (list): list of material ids

        Returns: (train_ids, val_ids, test_ids) or
            (train_ids, test_ids)
        """


class ShuffleSplitter(DataSplitter):
    """
    Randomly shuffe the material ids and split the ids
    into given ratios
    """

    def __init__(self, ratios: str = "80/10/10", delim: str = "/", random_seed: Optional[int] = None):
        """
        Random shuffle split of material ids
        Args:
            ratios (str): ratios
            delim (str): deliminators for separate ratios
            random_seed (int): random seed
        """
        self.random_seed = random_seed
        int_ratios = [float(i) for i in ratios.strip().split(delim)]
        self.ratios = [i / sum(int_ratios) for i in int_ratios]

    def split(self, mat_ids, **kwargs) -> Tuple:
        """
        Randomly split the mat_ids
        Args:
            mat_ids (list): material ids
        Returns:

        """
        mat_id_copy = shuffle(mat_ids, random_state=self.random_seed)
        n = len(mat_ids)
        end_points = [0] + np.cumsum([int(n * i) for i in self.ratios[:-1]]).tolist() + [n]
        return tuple(mat_id_copy[i:j] for i, j in zip(end_points[:-1], end_points[1:]))



"""
Keras utils for deserialize activations and otpimizers
"""


def deserialize_keras_activation(activation):
    """
    serialize keras activation

    Args:
        activation (str, dict, or Activation object): activation to serialize

    Returns: Activation object
    """
    import tensorflow as tf

    if isinstance(activation, str):
        return tf.keras.activations.get(activation)

    if isinstance(activation, dict):
        return tf.keras.activations.deserialize(activation)

    if isinstance(activation, tf.keras.activations.Activation):
        return activation

    raise ValueError("activation not recognized")


def deserialize_keras_optimizer(optimizer):
    """
    serialize keras activation

    Args:
        optimizer (str, dict, or optimizer object): optimizer to serialize

    Returns: Activation object
    """
    import tensorflow as tf

    if isinstance(optimizer, str):
        return tf.keras.optimizers.get(optimizer)

    if isinstance(optimizer, dict):
        return tf.keras.optimizers.deserialize(optimizer)

    if isinstance(optimizer, tf.keras.optimizers.Optimizer):
        return optimizer

    raise ValueError("optimizer not recognized")



"""
Module implements helper functions to retrieve data for GB energy prediction paper
"""
import os
from typing import Dict, List, Optional

from monty.serialization import dumpfn, loadfn
from pymatgen.core import Structure
from pymatgen.ext.matproj import MPRester, MPRestError

from maml.apps import gbe

pjoin = os.path.join
module_dir = os.path.dirname(gbe.__file__)
REFS = pjoin(module_dir, "references")
DATA = pjoin(module_dir, "data")


def load_data(filename: Optional[str] = None) -> List:
    """
    Helper function to load the data
    Default is to load the 361 data
    Args:
        filename (str): the filename of the data

    Returns:
        data list

    """
    if filename:
        return loadfn(filename)
    return loadfn(pjoin(DATA, "gb_data.json"))


def load_b0_dict() -> Dict:
    """
    Helper function to retrieve the b0 data.
    b0 is the bulk bond length.

    Returns:
        returns the dict of {element: b0}
    """
    if os.path.isfile(pjoin(REFS, "el2b0.json")):
        return loadfn(pjoin(REFS, "el2b0.json"))
    print("Downloading Data ... ")
    if os.getenv("MAPI_KEY"):
        data = load_data()
        b0_dict = update_b0_dict(data, mp_api=os.getenv("MAPI_KEY"))
        dumpfn(b0_dict, pjoin(REFS, "el2b0.json"))
        return b0_dict
    raise MPRestError("No MAPI_KEY found")


def update_b0_dict(data_source: list, mp_api: Optional[str]) -> Dict:
    """
    Helper function to update the b0 dictionary
    Requires api key to MP
    Args:
        data_source (list): list of GB data
        mp_api (str): API key to MP
    Returns:
         b0_dict (dict): {el: b0}
    """
    rester = MPRester(mp_api)

    def get_b0(bulk_structure: Structure) -> float:
        """b0 is the bond length of bulk metal"""
        _b0 = bulk_structure.get_distance(0, 0, jimage=1)
        return min(_b0, bulk_structure.lattice.a)

    b0_dict = {}
    for d in data_source:
        if d["pretty_formula"] in b0_dict:
            continue
        bs = rester.get_structure_by_material_id(d["material_id"])
        b0 = get_b0(bs)
        b0_dict.update({d["pretty_formula"]: b0})
    return b0_dict


def load_mean_delta_bl_dict(loc_algo: str = "crystalnn") -> Dict:
    """
    Helper function to load the mean_delta_bl data
    Args:
        loc_algo (str): name of the algorithm

    Returns:
        {task_id (int): mean_delta_bl(float)}

    """
    if os.path.isfile(pjoin(REFS, f"mean_bl_chg_{loc_algo}.json")):
        return loadfn(pjoin(REFS, f"mean_bl_chg_{loc_algo}.json"))
    raise ValueError(
        "Please provide mean_delta_bond_length data. "
        "Use gbe.describer.GBond.get_mean_bl_chg method"
        "to calculate mean_delta_bond_length"
    )



"""
MAML data base classes
"""
import abc

import pandas as pd


class BaseDataSource:
    """
    Abstract base class for a data source.
    """

    @abc.abstractmethod
    def get(self, *args, **kwargs) -> pd.DataFrame:
        """
        Get data from sources
        """



"""
neural network models
"""
import json
import math
import os
from typing import Any, List, Optional, Sequence, Union

import joblib
import numpy as np

from maml.base import BaseDescriber, BaseModel, KerasModel

from ._keras_utils import deserialize_keras_activation, deserialize_keras_optimizer


def construct_atom_sets(
    input_dim: Optional[int] = None,
    is_embedding: bool = True,
    n_neurons: Sequence[int] = (64, 64),
    n_neurons_final: Sequence[int] = (64, 64),
    n_targets: int = 1,
    activation: str = "relu",
    embedding_vcal: int = 95,
    embedding_dim: int = 32,
    symmetry_func: Union[List[str], str] = "mean",
    optimizer: str = "adam",
    loss: str = "mse",
    compile_metrics: tuple = (),
    is_classification: bool = False,
    **symmetry_func_kwargs,
):
    r"""
    f(X) = \rho(\sum_{x \in X} \phi(x)), where X is a set.
    \phi is implemented as a neural network and \rho is a symmetry function.

    todo: implement attention mechanism

    Args:
        input_dim (int): input dimension, if None, then integer inputs + embedding are assumed.
        is_embedding (bool): whether the input should be embedded
        n_neurons (tuple): number of hidden-layer neurons before passing to symmetry function
        n_neurons_final (tuple): number of hidden-layer neurons after symmetry function
        n_targets (int): number of output targets
        activation (str): activation function
        embedding_vcal (int): int, embedding vocabulary
        embedding_dim (int): int, embedding dimension
        symmetry_func (str): symmetry function, choose from ['set2set', 'sum', 'mean',
            'max', 'min', 'prod']
        optimizer (str): optimizer for the models
        loss (str): loss function for the models
        compile_metrics (tuple): metrics for validation
        symmetry_func_kwargs (dict): kwargs for symmetry function
    """
    from tensorflow.keras.layers import Concatenate, Dense, Embedding, Input
    from tensorflow.keras.models import Model

    if is_embedding and input_dim is not None:
        raise ValueError("When embedding is used, input dim needs to be None")

    if is_embedding:
        inp = Input(shape=(None,), dtype="int32", name="node_id")
        out_ = Embedding(embedding_vcal, embedding_dim)(inp)
    else:
        inp = Input(shape=(None, input_dim), dtype="float32", name="node_feature_input")
        out_ = inp

    weight_inputs = Input(shape=(None,), dtype="float32", name="weight_input")
    node_ids = Input(shape=(None,), dtype="int32", name="node_in_graph_id")

    # start neural networks \phi
    for n_neuron in n_neurons:
        out_ = Dense(n_neuron, activation=activation)(out_)

    # apply symmetry function \rho
    if isinstance(symmetry_func, str):
        symmetry_func = [symmetry_func]

    symmetry_layers = []
    for symm in symmetry_func:
        if symm == "set2set":
            from maml.models.dl._layers import WeightedSet2Set

            layer = WeightedSet2Set(**symmetry_func_kwargs)
        elif symm == "mean":
            from maml.models.dl._layers import WeightedAverageLayer

            alpha = symmetry_func_kwargs.pop("alpha", 1)
            layer = WeightedAverageLayer(alpha=alpha)
        else:
            raise ValueError("symmetry function not supported")
        symmetry_layers.append(layer)
    outs = [i([out_, weight_inputs, node_ids]) for i in symmetry_layers]

    if len(outs) > 1:
        out_ = Concatenate(axis=-1)(outs)
    else:
        out_ = outs[0]

    # neural networks
    for n_neuron in n_neurons_final:
        out_ = Dense(n_neuron, activation=activation)(out_)

    if is_classification:
        final_act: Optional[str] = "sigmoid"
    else:
        final_act = None
    out_ = Dense(n_targets, activation=final_act)(out_)
    model = Model(inputs=[inp, weight_inputs, node_ids], outputs=out_)
    model.compile(optimizer, loss, metrics=compile_metrics)
    return model


class AtomSets(KerasModel):
    r"""
    This class implements the DeepSets models
    """

    def __init__(
        self,
        describer: Optional[BaseDescriber] = None,
        input_dim: Optional[int] = None,
        is_embedding: bool = True,
        n_neurons: Sequence[int] = (64, 64),
        n_neurons_final: Sequence[int] = (64, 64),
        n_targets: int = 1,
        activation: str = "relu",
        embedding_vcal: int = 95,
        embedding_dim: int = 32,
        symmetry_func: Union[List[str], str] = "mean",
        optimizer: str = "adam",
        loss: str = "mse",
        compile_metrics: tuple = (),
        is_classification: bool = False,
        **symmetry_func_kwargs,
    ):
        """
        Args:
            describer (BaseDescriber): site describers
            input_dim (int): input dimension, if None, then integer inputs + embedding are assumed.
            is_embedding (bool): whether the input should be embedded
            n_neurons (tuple): number of hidden-layer neurons before passing to symmetry function
            n_neurons_final (tuple): number of hidden-layer neurons after symmetry function
            n_targets (int): number of output targets
            activation (str): activation function
            embedding_vcal (int): int, embedding vocabulary
            embedding_dim (int): int, embedding dimension
            symmetry_func (str): symmetry function, choose from ['set2set', 'sum', 'mean',
                'max', 'min', 'prod']
            optimizer (str): optimizer for the models
            loss (str): loss function for the models
            symmetry_func_kwargs (dict): kwargs for symmetry function

        """

        input_dim = self.get_input_dim(describer, input_dim)

        optimizer = deserialize_keras_optimizer(optimizer)
        activation = deserialize_keras_activation(activation)

        model = construct_atom_sets(
            input_dim=input_dim,
            is_embedding=is_embedding,
            n_neurons=n_neurons,
            n_neurons_final=n_neurons_final,
            n_targets=n_targets,
            activation=activation,
            embedding_vcal=embedding_vcal,
            embedding_dim=embedding_dim,
            symmetry_func=symmetry_func,
            optimizer=optimizer,
            loss=loss,
            compile_metrics=compile_metrics,
            is_classification=is_classification,
            **symmetry_func_kwargs,
        )

        self.input_dim = input_dim
        self.is_embedding = is_embedding
        self.n_neurons = n_neurons
        self.n_neurons_final = n_neurons_final
        self.n_targets = n_targets
        self.activation = activation
        self.embedding_vcal = embedding_vcal
        self.embedding_dim = embedding_dim
        self.symmetry_func = symmetry_func
        self.optimizer = optimizer
        self.loss = loss
        self.compile_metrics = compile_metrics
        self.is_classification = is_classification
        self.symmetry_func_kwargs = symmetry_func_kwargs
        super().__init__(model=model, describer=describer)

    def _get_data_generator(self, features, targets, batch_size=128, is_shuffle=True):
        if features is None:
            return None
        from tensorflow.keras.utils import Sequence as KerasSequence

        def _generate_atom_indices(lengths):
            max_length = max(lengths)
            res = np.tile(np.arange(max_length)[None, :], (len(lengths), 1))
            res2 = np.tile(np.arange(len(lengths))[:, None], (1, max_length))
            return res2[res < np.array(lengths)[:, None]]

        is_embedding = self.is_embedding

        class _DataGenerator(KerasSequence):
            def __init__(self, features=features, targets=targets, batch_size=batch_size, is_shuffle=is_shuffle):
                if isinstance(features[0], list) and len(features[0]) == 2:
                    self.features = [i[0] for i in features]
                    self.weights = [i[1] for i in features]
                else:
                    self.features = features
                    self.weights = [np.ones(shape=(len(i),)) for i in features]
                self.targets = targets
                self.batch_size = batch_size
                self.is_shuffle = is_shuffle

            def __len__(self):
                return math.ceil(len(self.features) / self.batch_size)

            def __getitem__(self, idx):
                features_temp = self.features[idx * self.batch_size : (idx + 1) * self.batch_size]
                batch_x = np.concatenate(features_temp, axis=0)
                if is_embedding:
                    batch_x = batch_x[..., 0]
                lengths = [len(i) for i in features_temp]
                f_index = _generate_atom_indices(lengths)
                batch_weights = self.weights[idx * self.batch_size : (idx + 1) * self.batch_size]
                batch_y = self.targets[idx * self.batch_size : (idx + 1) * self.batch_size]
                return (
                    np.array(batch_x)[None, ...],
                    np.concatenate(batch_weights)[None, :],
                    f_index[None, :],
                ), np.array(batch_y)[None, :]

            def on_epoch_end(self):
                """
                Codes executed at the end of each epoch
                """
                if self.is_shuffle:
                    indices = list(range(len(self.features)))
                    np.random.shuffle(indices)
                    self.features = [self.features[i] for i in indices]
                    self.weights = [self.weights[i] for i in indices]
                    self.targets = [self.targets[i] for i in indices]

        return _DataGenerator()

    def save(self, dirname: str):
        """Save the models and describers

        Arguments:
            dirname (str): dirname for save
        """
        import tensorflow as tf

        if not os.path.isdir(dirname):
            os.mkdir(dirname)

        joblib.dump(self.describer, os.path.join(dirname, "describer.sav"))

        def _to_json_type(d):
            if isinstance(d, dict):
                return {i: _to_json_type(j) for i, j in d.items()}
            if isinstance(d, np.float32):
                return float(d)
            if isinstance(d, np.int32):
                return int(d)
            if isinstance(d, (list, tuple)):
                return [_to_json_type(i) for i in d]
            return d

        with open(os.path.join(dirname, "config.json"), "w") as f:
            json.dump(
                {
                    "input_dim": self.input_dim,
                    "is_embedding": self.is_embedding,
                    "n_neurons": self.n_neurons,
                    "n_neurons_final": self.n_neurons_final,
                    "n_targets": self.n_targets,
                    "activation": _to_json_type(tf.keras.activations.serialize(self.activation)),
                    "embedding_vcal": self.embedding_vcal,
                    "embedding_dim": self.embedding_dim,
                    "symmetry_func": self.symmetry_func,
                    "optimizer": _to_json_type(tf.keras.optimizers.serialize(self.optimizer)),
                    "loss": self.loss,
                    "compile_metrics": self.compile_metrics,
                    "is_classification": self.is_classification,
                    "symmetry_func_kwargs": self.symmetry_func_kwargs,
                },
                f,
            )
        self.model.save(os.path.join(dirname, "model_weights.hdf5"))

    @classmethod
    def from_dir(cls, dirname: str):
        """
        Load the models from file
        Args:
            dirname (str): directory name
        Returns: object instance
        """
        with open(os.path.join(dirname, "config.json")) as f:
            kwarg_dict = json.load(f)

        symmetry_kwargs = kwarg_dict.pop("symmetry_func_kwargs")
        kwarg_dict.update(**symmetry_kwargs)

        describer = joblib.load(os.path.join(dirname, "describer.sav"))
        instance = cls(describer=describer, **kwarg_dict)
        instance.model.load_weights(os.path.join(dirname, "model_weights.hdf5"))
        return instance

    def fit(
        self,
        features: Union[List, np.ndarray],
        targets: Union[List, np.ndarray] = None,
        val_features: Optional[Union[List, np.ndarray]] = None,
        val_targets: Optional[Union[List, np.ndarray]] = None,
        **kwargs,
    ) -> "BaseModel":
        """
        Args:
            features (list or np.ndarray): Numerical input feature list or
                numpy array with dim (m, n) where m is the number of data and
                n is the feature dimension.
            targets (list or np.ndarray): Numerical output target list, or
                numpy array with dim (m, ).
            val_features (list or np.ndarray): validation features
            val_targets (list or np.ndarray): validation targets

        """
        batch_size = kwargs.pop("batch_size", 128)
        is_shuffle = kwargs.pop("is_shuffle", True)
        train_generator = self._get_data_generator(features, targets, batch_size=batch_size, is_shuffle=is_shuffle)
        val_generator = self._get_data_generator(
            val_features, val_targets, batch_size=batch_size, is_shuffle=is_shuffle
        )
        return self.model.fit(train_generator, validation_data=val_generator, **kwargs)  # type: ignore

    def _predict(self, features: np.ndarray, **kwargs) -> np.ndarray:
        """
        Predict the values given a set of inputs based on fitted models.

        Args:
            features (np.ndarray): array-like input features.

        Returns:
            List of output objects.
        """
        predict_generator = self._get_data_generator(features, [0] * len(features), is_shuffle=False, **kwargs)
        predicted = []
        for batch in predict_generator:
            predicted.append(self.model.predict(batch[0])[0])  # type: ignore
        return np.concatenate(predicted, axis=0)

    def evaluate(self, eval_objs: Any, eval_targets: Any, is_feature: bool = False, batch_size: int = 16):
        """
        Evaluate objs, targets

        Args:
            eval_objs (list): objs for evaluation
            eval_targets (list): target list for the corresponding objects
            is_feature (bool): whether x is feature matrix
            batch_size (int): evaluation batch size
        """
        eval_features = eval_objs if is_feature else self.describer.transform(eval_objs)
        eval_generator = self._get_data_generator(eval_features, eval_targets, is_shuffle=False, batch_size=batch_size)
        return self.model.evaluate(eval_generator)



"""CGCNN Wrapper."""
import argparse
import os
import warnings
from typing import Any, Dict, List, Tuple

import numpy as np
from monty.dev import requires
from pymatgen.core.structure import Structure

from maml.apps.bowsr.model.base import EnergyModel

try:
    import cgcnn
    import torch
    from cgcnn.data import AtomCustomJSONInitializer, GaussianDistance
    from cgcnn.model import CrystalGraphConvNet
    from torch import Tensor
except ImportError:
    torch = None  # type: ignore
    cgcnn = None  # type: ignore
    Tensor = None  # type: ignore

pjoin = os.path.join
module_dir = os.path.dirname(__file__)
model_dir = pjoin(module_dir, "model_files", "cgcnn", "formation-energy-per-atom.pth.tar")


@requires(cgcnn is not None and torch is not None, "cgcnn and torch are needed to use the CGCNN evaluator.")
class CGCNN(EnergyModel):
    """Wrapper to generate cgcnn energy prediction model."""

    def __init__(self, model_path: str = model_dir, orig_atom_fea_len: int = 92, nbr_fea_len: int = 41):
        """
        Init CGCNN.
        Args:
            model_path(str): path of model
            orig_atom_fea_len(int): Number of atom features in the input.
                                    i.e. Original atom feature length
                                    (default 92)
            nbr_fea_len(int): Number of bond features.
                            i.e. Number of neighbors (default 41)

        """
        checkpoint = torch.load(model_path, map_location=lambda storage, loc: storage)
        model_args = argparse.Namespace(**checkpoint["args"])
        self.model = CrystalGraphConvNet(
            orig_atom_fea_len=orig_atom_fea_len,
            nbr_fea_len=nbr_fea_len,
            n_h=model_args.n_h,
            n_conv=model_args.n_conv,
            h_fea_len=model_args.h_fea_len,
        )
        self.normalizer = CGCNNNormalizer(torch.zeros(3))
        self.model.load_state_dict(checkpoint["state_dict"])
        self.normalizer.load_state_dict(checkpoint["normalizer"])

    def predict_energy(self, structure: Structure) -> np.ndarray:  # type: ignore
        """
        CGCNN predict formatio nenergy from pymatgen structure.

        Args:
            structure(Structure): structure to be predicted

        Returns: formation energy (eV/atom) of provided structure

        """
        self.model.eval()
        input_generator = CGCNNInput()
        inp = input_generator.generate_input(structure)
        inp = inp + ([torch.LongTensor(np.arange(structure.num_sites))],)  # type: ignore
        output = self.model(*inp)
        return self.normalizer.denorm(output).data.cpu().numpy()[0][0]


@requires(cgcnn is not None and torch is not None, "cgcnn and torch are needed to use the CGCNN evaluator.")
class CGCNNInput:
    """Wrapper to generate input for cgcnn from pymatgen structure."""

    atom_init_filename = pjoin(module_dir, "model_files", "cgcnn", "atom_init.json")

    def __init__(self, max_num_nbr: int = 12, radius: float = 8, dmin: float = 0, step: float = 0.2):

        """
        Init CGCNNInput.

        Args:
            max_num_nbr(int): The maximum number of neighbors
                            while constructing the crystal graph
                            (default 12)
            radius(float): The cutoff radius for searching neighbors
                            (default 8)
            dmin(float): The minimum distance for constructing
                        GaussianDistance (default 0)
            step(float): The step size for constructing GaussianDistance
                        (default 0.2)
        """
        self.max_num_nbr = max_num_nbr
        self.radius = radius
        self.dmin = dmin
        self.step = step
        self.gdf = GaussianDistance(dmin=dmin, dmax=self.radius, step=step)
        self.ari = AtomCustomJSONInitializer(self.atom_init_filename)

    def _get_nbr_fea(self, all_nbrs: list, cif_id: int) -> Tuple[np.ndarray, ...]:
        nbr_fea_idx, nbr_fea = [], []
        for nbr in all_nbrs:
            if len(nbr) < self.max_num_nbr:
                warnings.warn(
                    f"{cif_id} not find enough neighbors to build graph. "
                    "If it happens frequently, consider increase "
                    "radius."
                )
                nbr_fea_idx.append(list(map(lambda x: x[2], nbr)) + [0] * (self.max_num_nbr - len(nbr)))
                nbr_fea.append(list(map(lambda x: x[1], nbr)) + [self.radius + 1.0] * (self.max_num_nbr - len(nbr)))
            else:
                nbr_fea_idx.append(list(map(lambda x: x[2], nbr[: self.max_num_nbr])))
                nbr_fea.append(list(map(lambda x: x[1], nbr[: self.max_num_nbr])))
        nbr_fea_idx, nbr_fea = np.array(nbr_fea_idx), np.array(nbr_fea)  # type: ignore
        nbr_fea = self.gdf.expand(nbr_fea)
        return tuple((nbr_fea_idx, nbr_fea))  # type: ignore

    def generate_input(self, structure: Structure, cif_id: int = None) -> Tuple[Any, ...]:

        """
        Generate cgcnn inputs for given structure.

        Args:
            structure(Structure): structure to get input for
            cif_id(int): Optional, the id of the structure

        Returns: Tuple of input (atom_fea, nbr_fea, nbr_fea_idx)
        """
        atom_fea = [sum((self.ari.get_atom_fea(el.Z) * oc) for el, oc in site.species.items()) for site in structure]
        atom_fea = np.vstack(atom_fea)  # type: ignore
        atom_fea = Tensor(atom_fea)  # type: ignore
        all_nbrs = structure.get_all_neighbors(self.radius, include_index=True)
        # sort the nbrs by distance
        all_nbrs = [sorted(nbrs, key=lambda x: x[1]) for nbrs in all_nbrs]

        if cif_id:
            nbr_fea_idx, nbr_fea = self._get_nbr_fea(all_nbrs, cif_id)
        else:
            nbr_fea_idx, nbr_fea = self._get_nbr_fea(all_nbrs, 0)
        nbr_fea = Tensor(nbr_fea)  # type: ignore
        nbr_fea_idx = torch.LongTensor(nbr_fea_idx)  # type: ignore
        return tuple((atom_fea, nbr_fea, nbr_fea_idx))

    def generate_inputs(self, structures: List[Structure], cif_ids: List[int] = None) -> List[Tuple[Any, ...]]:
        """
        Generate cgcnn inputs for given list of structures
        Args:
            structures (list): List of structures to get inputs for.
            cif_ids (list): Optional, the list of ids of the structures.

        """
        if not cif_ids:
            cif_ids = list(range(len(structures)))
        return [self.generate_input(s, id) for s, id in zip(structures, cif_ids)]


@requires(cgcnn is not None and torch is not None, "cgcnn and torch are needed to use the CGCNN evaluator.")
class CGCNNNormalizer:
    """Normalize a Tensor and restore it later."""

    def __init__(self, tensor: Tensor):
        """
        Tensor is taken as a sample to calculate the mean and std.

        Args:
            tensor(Tensor): data
        """
        self.mean = torch.mean(tensor)
        self.std = torch.std(tensor)

    def norm(self, tensor: Tensor) -> Tensor:
        """
        Normalize tensor.

        Args:
            tensor(Tensor): data

        Returns: normalized tensor

        """
        return (tensor - self.mean) / self.std

    def denorm(self, normed_tensor: Tensor) -> Tensor:
        """
        Denormalize tensor.

        Args:
            normed_tensor(Tensor): normalized tensor data

        Returns: denormalized tensor

        """
        return normed_tensor * self.std + self.mean

    def state_dict(self) -> dict:
        """
        Get dict of mean and std.

        Returns: dict of mean and std of the normalizer。

        """
        return {"mean": self.mean, "std": self.std}

    def load_state_dict(self, state_dict: Dict) -> None:
        """
        Load the normalizer with mean and std.

        Args:
            state_dict(Dict): dict of mean and std

        Returns: None

        """
        self.mean = state_dict["mean"]
        self.std = state_dict["std"]



"""
Module implements the describer for GB entry
"""
from functools import reduce
from math import gcd
from typing import Dict

import numpy as np
import pandas as pd
from monty.json import MSONable
from pymatgen.analysis.gb.grain import GrainBoundary
from pymatgen.analysis.local_env import (
    BrunnerNN_real,
    BrunnerNN_reciprocal,
    BrunnerNN_relative,
    CovalentBondNN,
    Critic2NN,
    CrystalNN,
    CutOffDictNN,
    EconNN,
    JmolNN,
    MinimumDistanceNN,
    MinimumOKeeffeNN,
    MinimumVIRENN,
    NearNeighbors,
    OpenBabelNN,
    VoronoiNN,
)
from pymatgen.core import Element, Structure
from pymatgen.ext.matproj import MPRester, MPRestError
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

import maml.apps.gbe.presetfeatures as preset
from maml.apps.gbe.utils import load_b0_dict, load_mean_delta_bl_dict
from maml.base import BaseDescriber


def convert_hcp_direction(rotation_axis: list, lat_type: str) -> np.ndarray:
    """
    four index notion to three index notion for hcp and rhombohedral axis
    Args:
        rotation_axis (list): four index notion of axis
        lat_type(str): the
    Returns:
        rotation axis in three index notion
    """
    u1 = rotation_axis[0]
    v1 = rotation_axis[1]
    w1 = rotation_axis[3]
    if lat_type.lower().startswith("h"):
        u = 2 * u1 + v1
        v = 2 * v1 + u1
        w = w1
        rotation_axis = [u, v, w]
    elif lat_type.lower().startswith("r"):
        u = 2 * u1 + v1 + w1
        v = v1 + w1 - u1
        w = w1 - 2 * v1 - u1
        rotation_axis = [u, v, w]

    # make sure gcd(rotation_axis)==1
    if reduce(gcd, rotation_axis) != 1:
        rotation_axis = [int(round(x / reduce(gcd, rotation_axis))) for x in rotation_axis]
    return np.array(rotation_axis)


def convert_hcp_plane(plane: list) -> np.ndarray:
    """
    four index notion to three index notion for hcp and rhombohedral plane
    Args:
        plane (list):  four index notion

    Returns:
        three index notion of plane

    """
    u1 = plane[0]
    v1 = plane[1]
    w1 = plane[3]
    plane = [u1, v1, w1]
    if reduce(gcd, plane) != 1:
        index = reduce(gcd, plane)
        plane = [int(round(x / index)) for x in plane]
    return np.array(plane)


class GBDescriber(BaseDescriber):
    """
    The describer that descibe the grain boundary db entry
    with selected structural and elemental features
    """

    def __init__(self, structural_features: list = None, elemental_features: list = None, **kwargs):
        """

        Args:
            structural_features (list): list of structural features
            elemental_features (list): list of elemental features
            **kwargs (dict): parameters for BaseDescriber
        """
        if not elemental_features:
            elemental_features = [preset.e_coh, preset.G, preset.a0, preset.ar, preset.mean_delta_bl, preset.mean_bl]
        if not structural_features:
            structural_features = [preset.d_gb, preset.d_rot, preset.sin_theta, preset.cos_theta]
        self.elem_features = elemental_features
        self.struc_features = structural_features
        super().__init__(**kwargs)

    def transform_one(
        self, db_entry: Dict, inc_target: bool = True, inc_bulk_ref: bool = True, mp_api: str = None
    ) -> pd.DataFrame:
        """
        Describe gb with selected structural and elemental features
        Args:
            db_entry (dict): entry from surfacedb.All_GB_properties_parallel_copy
            inc_target (bool): whether to include target in the df, default: False
            inc_bulk_ref (bool): whether to generate bulk reference
                bulk reference: i.e. the entry of the origin bulk of the GB,
                                the rotation angle (theta) = 0, gb_energy = 0
            mp_api (str): MP api key


        Returns: pd.DataFrame of processed data, columns are the feature labels
                 The task_id is included and serve as the unique id of the data

        """
        structural = get_structural_feature(db_entry=db_entry, features=self.struc_features)
        elemental = get_elemental_feature(db_entry=db_entry, features=self.elem_features, mp_api=mp_api)
        df = pd.concat([structural, elemental], axis=1, join="inner")
        df["task_id"] = db_entry["task_id"]
        if inc_target:
            df[preset.e_gb.str_name] = db_entry["gb_energy"]
        if inc_bulk_ref:
            df_bulk = self.generate_bulk_ref(df, inc_target)
            return pd.concat([df, df_bulk], axis=0)
        return df

    def generate_bulk_ref(self, gb_df: pd.DataFrame, inc_target: bool = True) -> pd.DataFrame:
        """
        Generate the bulk reference for given gb entry
        Args:
            gb_df (pd.DataFrame): data for gb
            inc_target (bool) : whether to include target

        Returns:
            bulk_df (pd.DataFrame): data for bulk

        """
        new_df = gb_df.copy()
        new_df[preset.sin_theta.str_name] = np.sin(0)
        new_df[preset.cos_theta.str_name] = np.cos(0)
        if preset.mean_delta_bl.str_name in new_df:
            new_df[preset.mean_delta_bl.str_name] = 0
        if inc_target:
            new_df[preset.e_gb.str_name] = 0
        return new_df


def get_structural_feature(db_entry: Dict, features: list = None) -> pd.DataFrame:
    """
    The structural features:
    d_gb: interplanal distance of the gb_plane
    d_rot: interplanal distance of the gb_plane
        w/ smallest interger index ) normal to rotation axis
    theta: rotation angle (sin and cos)
    Args:
        db_entry (Dict): db entry
        features (List): list of features


    Returns:
        pd.DataFrame of structural features

    """
    if features is None:
        features = [preset.d_gb, preset.d_rot, preset.sin_theta, preset.cos_theta]
    if isinstance(db_entry["bulk_conv"], dict):
        bulk_conv = Structure.from_dict(db_entry["bulk_conv"])
    else:
        bulk_conv = db_entry["bulk_conv"]
    sg = SpacegroupAnalyzer(bulk_conv)
    gb_plane = np.array(db_entry["gb_plane"])
    rotation_axis = db_entry["rotation_axis"]
    if gb_plane.shape[0] == 4:
        gb_plane = convert_hcp_plane(list(gb_plane))
        rotation_axis = convert_hcp_direction(rotation_axis, lat_type=sg.get_crystal_system())
    d_gb = bulk_conv.lattice.d_hkl(gb_plane)
    d_rot = bulk_conv.lattice.d_hkl(rotation_axis)
    theta = db_entry["rotation_angle"]
    sin_theta = np.sin(theta * np.pi / 180)
    cos_theta = np.cos(theta * np.pi / 180)
    fvalues = []
    for f in features:
        if f == preset.sin_theta:
            fvalues.append(np.sin(theta * np.pi / 180))
        elif f == preset.cos_theta:
            fvalues.append(np.cos(theta * np.pi / 180))
        else:
            fvalues.append(locals()[f"{f.str_name}"])
    return pd.DataFrame([fvalues], columns=[f.str_name for f in features])


def get_elemental_feature(
    db_entry: Dict, loc_algo: str = "crystalnn", features: list = None, mp_api: str = None
) -> pd.DataFrame:
    """
    Function to get the elemental features
    Args:
        db_entry (Dict): db entry
        loc_algo(str): algorithm to determine local env.
                    Options: see GBBond.NNDict.keys()
        features(List): list of feature names
                e_coh: cohesive energy
                G: G_vrh shear modulus
                a0: bulk lattice parameter a
                ar: atomic radius
                b0: the bond length of the metal bulk
                mean_delta_bl: the mean bond length difference
                    between GB and the bulk
                bl: the mean bond length in GB
        mp_api(str): api key to MP

    Returns:
        pd.DataFrame of elemental features
    """
    if features is None:
        features = [
            preset.e_coh,
            preset.G,
            preset.a0,
            preset.ar,
            preset.mean_delta_bl,
            preset.bdensity,
            preset.CLTE,
            preset.hb,
        ]
    if preset.mean_delta_bl in features or preset.mean_bl in features:
        mdbl = load_mean_delta_bl_dict(loc_algo=loc_algo)
    f_dict = {}
    if mp_api:
        rester = MPRester(mp_api)
    else:
        raise MPRestError("Please provide API key to access Materials Project")
    bulk = rester.get_data(db_entry["material_id"])
    bulk_s = rester.get_structure_by_material_id(db_entry["material_id"])
    if bulk:
        f_dict[preset.bdensity.str_name] = bulk[0]["density"]
        f_dict[preset.G.str_name] = bulk[0]["elasticity"]["G_VRH"]
    el = Element(db_entry["pretty_formula"])
    f_dict[preset.ar.str_name] = el.atomic_radius
    f_dict[preset.a0.str_name] = bulk_s.lattice.a
    f_dict[preset.e_coh.str_name] = rester.get_cohesive_energy(db_entry["material_id"])
    f_dict[preset.hb.str_name] = el.brinell_hardness
    f_dict[preset.CLTE.str_name] = el.coefficient_of_linear_thermal_expansion

    def get_mean_delta_bl(db_entry):
        return mdbl[str(db_entry["task_id"])]

    def get_bl(db_entry):
        b0_dict = load_b0_dict()
        b0 = b0_dict[db_entry["pretty_formula"]]
        return mdbl[str(db_entry["task_id"])] + b0

    fvalues = []
    for f in features:
        if f == preset.mean_delta_bl:
            # func = 'get_{}'.format(f.str_name).lower()
            # fvalues.append(locals()[func](db_entry))
            fvalues.append(get_mean_delta_bl(db_entry))
        elif f == preset.mean_bl:
            fvalues.append(get_bl(db_entry))
        else:
            fvalues.append(f_dict[f.str_name])
    return pd.DataFrame([fvalues], columns=[f.str_name for f in features])


class GBBond(MSONable):
    """
    This class describes the GB Bonding environment using provided
    local environment algorithm.
    Available algorithms: GBBond.NNDict.keys(), default CrystalNN
    """

    NNDict = {
        i.__name__.lower(): i
        for i in [
            NearNeighbors,
            VoronoiNN,
            JmolNN,
            MinimumDistanceNN,
            OpenBabelNN,
            CovalentBondNN,
            MinimumVIRENN,
            MinimumOKeeffeNN,
            BrunnerNN_reciprocal,
            BrunnerNN_real,
            BrunnerNN_relative,
            EconNN,
            CrystalNN,
            CutOffDictNN,
            Critic2NN,
        ]
    }

    def __init__(self, gb: GrainBoundary, loc_algo: str = "crystalnn", bond_mat: np.ndarray = None, **kwargs):
        """

        Args:
            gb (GrainBoundary): the GrainBoundary Object
            loc_algo (str): algorithm to determine local env.
                    See options: GBBond.NNDict.keys()
                    Default: crystalnn
            bond_mat (np.ndarray): optional

        """
        self.loc_algo = self.NNDict[loc_algo](**kwargs)
        self.gb = gb
        if isinstance(bond_mat, np.ndarray):
            self.bond_mat = bond_mat
        else:
            self.bond_mat = self._get_bond_mat(gb)

    def _get_bond_mat(self, gb: GrainBoundary) -> np.ndarray:
        """

        Args:
            gb (GrainBoundary): the grain boundary structure object

        Returns:
            bond matrix
            matrix of bond lengths
            1. bm[i][j] = bond length between atom i&j
            if i & j is bonded (determined by the loc_algo)
            2. bm[i][j] = bm[j][i]
            3. If not bonded, the bm[i][j] = np.nan

        """
        if isinstance(gb, dict):
            gb = GrainBoundary.from_dict(gb)
        bond_mat = np.zeros((gb.num_sites, gb.num_sites))
        all_nn_info = self.loc_algo.get_all_nn_info(gb)
        for i, _ in enumerate(gb.sites):
            # compute the all_nn_info first then call
            # the protected function is faster for multiple computation
            nn = self.loc_algo._get_nn_shell_info(structure=gb, site_idx=i, shell=1, all_nn_info=all_nn_info)
            for b in nn:
                bl = gb.get_distance(i, b["site_index"])
                bond_mat[i][b["site_index"]] = bl
        return bond_mat

    def get_breakbond_ratio(self, gb: GrainBoundary, b0: float, return_count: bool = False):
        """
        Get the breakbond ratio, i.e the ratio of shorter bonds vs longer bonds
        compared to the bulk bond length (b0)
        The algo to find the bonds can vary
        Meme: if use get_neighbors, require a hard set cutoff, which adds to
            arbitrariness

        Args:
            gb (GrainBoundary): a GrainBoundary object
            b0 (float): cutoff to determine short vs. long bonds,
                        default the bulk bond length
            return_count(bool): whether to return count of

        Returns:
            ratio (float): shorter_bond / longer_bonds
            if return_count:
            shorter_bond: # of short bonds
            longer_bond: # of long bonds
        """

        if not self.bond_mat:
            self._get_bond_mat(gb)
        long_bond = (self.bond_mat > b0).sum()
        short_bond = (self.bond_mat < b0).sum()
        ratio = short_bond / long_bond if long_bond else 0
        if return_count:
            return ratio, short_bond, long_bond
        return ratio

    @property
    def min_bl(self) -> float:
        """
        The minimum bond length
        """
        return self.bond_mat[self.bond_mat > 0].min()

    @property
    def max_bl(self) -> float:
        """
        The maximum bond length
        """
        return self.bond_mat.max()

    @property
    def bond_matrix(self) -> np.ndarray:
        """
        The (padded) bond matrix
        """
        return self.bond_mat

    def get_mean_bl_chg(self, b0: float) -> float:
        """
        Function to calculate the mean bond length difference between GB and the bulk
        Args:
            b0 (float): the bond length in bulk

        Returns: the mean_bl_chg

        """
        return (self.bond_mat[self.bond_mat > 0] - b0).mean()

    def as_dict(self) -> dict:
        """
        Dict representation of the GBond class

        Returns:
            dict of {"loc_algo": str, "bond_mat": bond matrix}
        """
        return_dict = super().as_dict()
        return_dict.update(
            {
                "loc_algo": self.loc_algo.__class__.__name__.lower(),
                "bond_mat": self.bond_mat,
                # "min_bl": self.min_bl,
                # "max_bl": self.max_bl
            }
        )
        return return_dict

    @classmethod
    def from_dict(cls, d):
        """

        Args:
            d (dict): Dict representation

        Returns:
            GBBond

        """
        return cls(gb=d["gb"], loc_algo=d["loc_algo"], bond_mat=d["bond_mat"])



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This module provides MTP interatomic potential class."""

import itertools
import json
import os
import re
import shutil
import subprocess
from collections import OrderedDict

import numpy as np
from monty.io import zopen
from monty.os.path import which
from monty.serialization import loadfn
from monty.tempfile import ScratchDir
from pymatgen.core import Lattice, Structure

from maml.utils import check_structures_forces_stresses, convert_docs, pool_from

from ._lammps import LammpsPotential

module_dir = os.path.dirname(__file__)
MTini_params = loadfn(os.path.join(module_dir, "params", "MTini.json"))


def feed(attribute, kwargs, dictionary, tab="\t"):
    r"""
    Args:
        attribute (str): Attribute to be operated.
        kwargs (dict): Generic parameters.
        dictionary (dict): The default parameters dictionary.
        tab (str): '\t' or '\t\t', depend on orders of attribute.
    Return:
        (str)
    """
    tmp = kwargs.get(attribute) if kwargs.get(attribute) else dictionary.get(attribute).get("value")
    return tab + dictionary.get(attribute).get("name"), str(tmp), dictionary.get(attribute).get("comment")


class MTPotential(LammpsPotential):
    """
    This class implements moment tensor potentials.
    Installation of the mlip package is needed.
    Please refer to https://mlip.skoltech.ru
    """

    pair_style = "pair_style        mlip {}"
    pair_coeff = "pair_coeff        * *"

    def __init__(self, name=None, param=None, version=None):
        """

        Args:
            name (str): Name of force field.
            param (dict): The parameter configuration of potentials.
            version (str): The version of mlip package. Default is "mlip-2". "mlip-dev" is also supported.
        """
        self.name = name if name else "MTPotential"
        self.mtp_stress_order = ["xx", "yy", "zz", "yz", "xz", "xy"]
        self.vasp_stress_order = ["xx", "yy", "zz", "xy", "yz", "xz"]
        self.param = param if param else None
        self.elements = None
        self.version = version if version else "mlip-2"

    def _line_up(self, structure, energy, forces, virial_stress):
        """
        Convert input structure, energy, forces, virial_stress to
        proper configuration format for mlip usage.

        Args:
            structure (Structure): Pymatgen Structure object.
            energy (float): DFT-calculated energy of the system.
            forces (list): The forces should have dimension (num_atoms, 3).
            virial_stress (list): stress should has 6 distinct
                elements arranged in order [xx, yy, zz, yz, xz, xy].
        """

        inputs = OrderedDict(
            Size=structure.num_sites,
            SuperCell=structure.lattice,
            AtomData=(structure, forces),
            Energy=energy,
            Stress=virial_stress,
        )

        lines = ["BEGIN_CFG"]

        if "Size" in inputs:
            lines.append(" Size")
            lines.append(f"{inputs['Size']:>7d}")
        if "SuperCell" in inputs:
            lines.append(" SuperCell")
            for vec in inputs["SuperCell"].matrix:
                lines.append(f"{vec[0]:>17.6f}{vec[1]:>14.6f}{vec[2]:>14.6f}")
        if "AtomData" in inputs:
            format_str = "{:>14s}{:>5s}{:>15s}{:>14s}{:>14s}{:>13s}{:>13s}{:>13s}"
            format_float = "{:>14d}{:>5d}{:>15f}{:>14f}{:>14f}{:>13f}{:>13f}{:>13f}"
            lines.append(
                format_str.format("AtomData:  id", "type", "cartes_x", "cartes_y", "cartes_z", "fx", "fy", "fz")
            )
            for i, (site, force) in enumerate(zip(structure, forces)):
                lines.append(format_float.format(i + 1, self.elements.index(str(site.specie)), *site.coords, *force))
        if "Energy" in inputs:
            lines.append(" Energy")
            lines.append(f"{inputs['Energy']:>24.12f}")
        if "Stress" in inputs:
            if not hasattr(self, "version") or self.version == "mlip-2":
                format_str = "{:>16s}{:>12s}{:>12s}{:>12s}{:>12s}{:>12s}"
                lines.append(format_str.format("PlusStress:  xx", "yy", "zz", "yz", "xz", "xy"))
            if self.version == "mlip-dev":
                format_str = "{:>12s}{:>12s}{:>12s}{:>12s}{:>12s}{:>12s}"
                lines.append(format_str.format("Stress:  xx", "yy", "zz", "yz", "xz", "xy"))
            format_float = "{:>12f}{:>12f}{:>12f}{:>12f}{:>12f}{:>12f}"
            lines.append(format_float.format(*np.array(virial_stress)))

        lines.append("END_CFG")

        return "\n".join(lines)

    def write_cfg(self, filename, cfg_pool):
        """
        Write configurations to file
        Args:
            filename (str): filename
            cfg_pool (list): list of configurations

        Returns:

        """
        if not self.elements:
            raise ValueError("No species given.")

        lines = []
        for dataset in cfg_pool:
            if isinstance(dataset["structure"], dict):
                structure = Structure.from_dict(dataset["structure"])
            else:
                structure = dataset["structure"]
            energy = dataset["outputs"]["energy"]
            forces = dataset["outputs"]["forces"]
            virial_stress = dataset["outputs"]["virial_stress"]
            virial_stress = [virial_stress[self.vasp_stress_order.index(n)] for n in self.mtp_stress_order]
            lines.append(self._line_up(structure, energy, forces, virial_stress))

        with open(filename, "w") as f:
            f.write("\n".join(lines))

        return filename

    def write_ini(self, mtp_filename="fitted.mtp", select=False, **kwargs):
        """
        Write mlip.ini file for mlip packages of version mlip-2 or mlip-dev.
        Supported keyword arguments are parallel with options stated in the mlip manuals.
        mlip-2 is recommended, as it is the only officially supported version by mlip.
        Please refer to https://mlip.skoltech.ru

        Args:
            mlip-2:
                mtp_filename (str): Name of file with MTP to be loaded.
                write_cfgs (str): Name of file for mlp processed configurations to be written to.
                write_cfgs_skip (int): Skipped number of processed configurations before writing.
                select (bool): activates or deactivates calculation of extrapolation grades and
                    optionally writing configurations with high extrapolation grades. False is
                    recommended for large-scale MD run.
                select_save_selected (str): Name of file for saving configurations with grade
                    exceeding select_threshold.
                select_threshold (float): Configurations with extrapolation grade exceeding ths
                    value will be saved to the specified file.
                select_threshold_break (float): The mlip executation will be interrupted if the
                    extrapolation grade exceeds this value.
                select_load_state (str): Name of file for loading the active learning state,
                    typically created by the mlp calc-grade command.
                select_log (str): Name of file (or standard output stream stdout/stderr) for
                    writing a log of the configuration selection process.

            mlip-dev:
                Abinitio (int): Defines Ab-initio models. Default to 1.

                    0: If Ab-initio models is not required.
                    1: Used if driver provides EFS data with configurations.
                    2: Use embedded Lennard-Jones pair potentials.

                        r_min (float): Distance to minimum of pair function (in Angstroms).
                            Default to 2.0.
                        scale (float): Value of pair function in minimum (in eV).
                            Default to 1.0.
                        cutoff (float): Cutoff radius (in Angstroms). Default to 5.0.

                    3: Use DFT models by VASP. Linking via files exchange.

                        POSCAR (str): Relative path of POSCAR file.
                        OUTCAR (str): Relative path of OUTCAR file.
                        Start_command (str): Relative path of command file.

                    4: Use potentials calculating by LAMMPS. Linking via files exchange.

                        Input_file (str): File with configuration to be read by lammps.
                        Output_file (str): File with configuration and EFS data to be read by MLIP.
                        Start_command (str): Relative path of command file.

                    5: Use MTP as Ab-initio potentials.

                        MTP_filename (str): MTP file name.

                MLIP (str): MTP.

                    load_from (str): Potential filename.
                    Cacluate_EFS (bool): Whether to perform EFS calculation by MTP.
                    Fit (bool): Whether to perform MTP learning.

                        Save (str): Output MTP file name (for trained MTP).
                        Energy_equation_weight (float): Weight for energy equation in
                            fitting procedure. Default to 1.0.
                        Forces_equation_weight (float): Weight for forces equations in
                            fitting procedure. Default to 0.001.
                        Stress_equation_weight (float): Weight for stresses equations in
                            fitting procedure.  Default to 0.1.
                        Relative_forces_weight (float): If greater than zero, large forces
                            will be fitted less accurate than small. Default to 0.0.
                        Fit_log (str): File to write fitting log. No logging if not specified.
                            Default to None.

                    Select (bool): Whether to activate active learning. Default to False.

                        Site_E_weight (float): Weight for site energy equations in
                            selection procedure. Default to 1.0.
                        Energy_weight (float): Weight for energy equation in
                            selection procedure. Default to 0.0.
                        Forces_weight (float): Weight for forces equations in
                            selection procedure. Default to 0.0.
                        Stress_weight (float): Weight for stresses equations in
                            selection procedure. Default to 0.0.
                        Threshold_slct (float): Selection threshold - maximum
                            allowed extrapolation level. Default to 0.1.
                        Save_TS (str): Filename where selected configurations
                            will be saved. No configuration saving if not specified.
                            Default to None.
                        Save_state (str): Filename where state of the selection
                            will be saved. No saving if not specified. Default to None.
                        Load_state (str): Filename where state of the selection
                            will be loaded. No saving if not specified. Default to None.
                        Select_log (str): File to write fitting log. No logging
                            if not specified. Default to None.

                    LOFT (bool): Whether to perform learning on the fly. Default to False

                        EFSviaMTP (bool): Works only on LOFT regime. If True,
                            only MTP-calculated EFS will be passed to driver, else
                            pass to driver ab-initio EFS while LOTF when learning occurs.
                        Log (str): Filename to write log of learning on the fly process.
                            No logging if not specified. Default to None.

                    Check_errors (bool): If True, comparison and accumulation of
                        error statistics for EFS calculated by ab-initio models and MTP.
                        Default to False.

                        Log (str): Filename to write log of learning on the fly process.
                            No logging if not specified. Default to None.

                    Write_cfgs (bool): File for writing all processed configurations.
                        No confuguration recording if not specified. Default to None.

                        Skip_N (int): The number of configurations to skip while writing.
                            Default to 0.

                    Log (str): Filename to write MLIP log. No logging if not specified.
                        Default to None.

                Driver (int): Defines the configuration driver. Default to 1.

                    0: No driver or external MD driver.
                    1: Read configurations from database file.

                        Database_filename (str): Configuration file name.
                        Max_count (int): Maximal number of configurations to read.
                        Log (str): Filename to write reading log. No logging
                            if not specified. Default to None.

                    2: Embedded algorithm for relaxation.

                        Pressure (float): External pressure (in GPa).
                            If not zero enthalpy is minimized. Default to 0.0.
                        Iteration_limit (int): Maximal number of iteration of
                            the relaxation process. Default to 500.
                        Min_dist (float): Minimal interatomic distance constraint
                            (in Angstroms). Default to 1.0.
                        Forces_tolerance (float): Forces on atoms in relaxed
                            configuration should be smaller than this value
                            (in eV/Angstroms). Default to 0.0001.
                        Stress_tolerance (float): Stresses in relaxed configuration
                            should be smaller than this value (in GPa). Default to 0.001.
                        Max_step (float): Maximal allowed displacement of atoms and
                            lattice vectors in Cartesian coordinates (in Angstroms).
                            Default to 0.5.
                        Min_step (float): Minimal displacement of atoms and
                            lattice vectors in Cartesian coordinates (in Angstroms).
                            Default to 1.0e-8.
                        BFGS_Wolfe_C1 (float): Wolfe condition constant on the function
                            decrease (linesearch stopping criterea). Default to 1.0e-3.
                        BFGS_Wolfe_C2 (float): Wolfe condition constant on the gradient
                            decrease (linesearch stopping criterea). Default to 0.7.
                        Save_relaxed (str): Filename for output results of relxation.
                            No configuration will be saved if not specified.
                            Default to None.
                        Log (str): Filename to write relaxation log. No logging
                            if not specified. Default to None.
        """
        lines = []
        if not hasattr(self, "version") or self.version == "mlip-2":
            format_str = "{:<48s}{:<20s}"
            lines.append(format_str.format("mtp-filename", mtp_filename))
            if kwargs.get("write_cfgs"):
                lines.append(format_str.format("write-cfgs", kwargs.get("write_cfgs")))
            if kwargs.get("write_cfgs_skip"):
                lines.append(format_str.format("write-cfgs:skip", kwargs.get("write_cfgs_skip")))
            if select is False:
                lines.append(format_str.format("select", "FALSE"))
            elif select is True:
                lines.append(format_str.format("select", "TRUE"))
                select_identifiers = [
                    "select:save-selected",
                    "select:threshold",
                    "select:threshold-break",
                    "select:load-state",
                    "select:log",
                ]
                for i, option in enumerate(
                    [
                        "select_save_selected",
                        "select_threshold",
                        "select_threshold_break",
                        "select_load_state",
                        "select_log",
                    ]
                ):
                    if kwargs.get(option):
                        lines.append(format_str.format("\t" + select_identifiers[i], kwargs.get(option)))

        elif self.version == "mlip-dev":
            format_str = "{:<48s}{:<20s}{}"
            Abinitio = kwargs.get("Abinitio") if kwargs.get("Abinitio") else 0
            MLIP = kwargs.get("MLIP") if kwargs.get("MLIP") else "MPT.mpt"
            Driver = kwargs.get("Driver") if kwargs.get("Driver") else 0
            PARAMS = {
                "Abinitio": {
                    0: [],
                    1: [],
                    2: ["r_min", "scale", "cutoff"],
                    3: ["POSCAR", "OUTCAR", "Start_command"],
                    4: ["Input_file", "Output_file", "Start_command"],
                    5: ["MTP_filename"],
                },
                "MLIP": {
                    "Calculate_EFS": [],
                    "Fit": [
                        "Save",
                        "Energy_equation_weight",
                        "Forces_equation_weight",
                        "Stress_equation_weight",
                        "Relative_forces_weight",
                        "Fit_log",
                    ],
                    "Select": [
                        "Site_E_weight",
                        "Energy_weight",
                        "Forces_weight",
                        "Stress_weight",
                        "Threshold_slct",
                        "Save_TS",
                        "Save_state",
                        "Load_state",
                        "Select_log",
                    ],
                    "Write_cfgs": [],
                    "Log": [],
                },
                "Driver": {
                    0: [],
                    1: ["Database_filename", "Database_log"],
                    2: [
                        "Pressure",
                        "Iteration_limit",
                        "Min_dist",
                        "Forces_tolerance",
                        "Stress_tolerance",
                        "Max_step",
                        "Min_step",
                        "BFGS_Wolfe_C1",
                        "BFGS_Wolfe_C2",
                        "Save_relaxed",
                        "Relaxation_log",
                    ],
                },
            }

            if Abinitio:
                lines.append(
                    format_str.format(
                        MTini_params.get("Abinitio").get("name"),
                        str(Abinitio),
                        MTini_params.get("Abinitio").get("comment"),
                    )
                )
                abinitio = MTini_params.get("Abinitio").get(str(Abinitio))
                lines.append(format_str.format(abinitio.get("name"), "", abinitio.get("comment")))
                for attribute in PARAMS["Abinitio"][Abinitio]:
                    lines.append(format_str.format(*feed(attribute, kwargs, abinitio)))

            if MLIP:
                lines.append(
                    format_str.format(
                        MTini_params.get("MLIP").get("name"), "mtpr", MTini_params.get("MLIP").get("comment")
                    )
                )
                mlip = MTini_params.get("MLIP")
                if kwargs.get("load_from"):
                    load_from = mlip.get("load_from")
                    lines.append(
                        format_str.format(
                            "\t" + load_from.get("name"), kwargs.get("load_from"), load_from.get("comment")
                        )
                    )
                if kwargs.get("Calculate_EFS"):
                    calc_efs = mlip.get("Calculate_EFS")
                    lines.append(format_str.format("\t" + calc_efs.get("name"), "TRUE", calc_efs.get("comment")))
                if kwargs.get("Fit"):
                    fit = mlip.get("Fit")
                    lines.append(format_str.format("\t" + fit.get("name"), "true", fit.get("comment")))
                    for attribute in PARAMS["MLIP"]["Fit"]:
                        lines.append(format_str.format(*feed(attribute, kwargs, fit, tab="\t\t")))

                if kwargs.get("Select"):
                    select = mlip.get("Select")
                    lines.append(format_str.format("\t" + select.get("name"), "true", select.get("comment")))
                    for attribute in PARAMS["MLIP"]["Select"]:
                        lines.append(format_str.format(*feed(attribute, kwargs, select, tab="\t\t")))

                if kwargs.get("Write_cfgs"):
                    write_cfgs = mlip.get("Write_cfgs")
                    lines.append(
                        format_str.format(
                            "\t" + write_cfgs.get("name"), kwargs.get("Write_cfgs"), write_cfgs.get("comment")
                        )
                    )

            if Driver:
                lines.append(
                    format_str.format(
                        MTini_params.get("Driver").get("name"), str(Driver), MTini_params.get("Driver").get("comment")
                    )
                )
                driver = MTini_params.get("Driver").get(str(Driver))
                lines.append(format_str.format(driver.get("name"), "", driver.get("comment")))
                for attribute in PARAMS["Driver"][Driver]:
                    lines.append(format_str.format(*feed(attribute, kwargs, driver, tab="\t\t")))

        filename = "mlip.ini"
        with open(filename, "w") as f:
            f.write("\n".join(lines))

        return filename

    def read_cfgs(self, filename):
        """

        Args:
            filename (str): The configuration file to be read.

        """

        def formatify(string):
            return [float(s) for s in string.split()]

        if not self.elements:
            raise ValueError("No species given.")

        data_pool = []
        with zopen(filename, "rt") as f:
            lines = f.read()

        block_pattern = re.compile("BEGIN_CFG\n(.*?)\nEND_CFG", re.S)
        size_pattern = re.compile("Size\n(.*?)\n SuperCell", re.S | re.I)
        lattice_pattern = re.compile("SuperCell\n(.*?)\n AtomData", re.S | re.I)
        position_pattern = re.compile("fz\n(.*?)\n Energy", re.S)
        energy_pattern = re.compile("Energy\n(.*?)\n (?=PlusStress|Stress)", re.S)
        stress_pattern = re.compile("xy\n(.*?)(?=\n|$)", re.S)
        for block in block_pattern.findall(lines):
            d = {"outputs": {}}
            size_str = size_pattern.findall(block)[0]
            size = int(size_str.lstrip())
            lattice_str = lattice_pattern.findall(block)[0]
            lattice = Lattice(np.array(list(map(formatify, lattice_str.split("\n")))))
            position_str = position_pattern.findall(block)[0]
            position = np.array(list(map(formatify, position_str.split("\n"))))
            species = np.array(self.elements)[position[:, 1].astype(np.int64)]
            forces = position[:, 5:8].tolist()
            position = position[:, 2:5]
            energy_str = energy_pattern.findall(block)[0]
            energy = float(energy_str.lstrip())
            stress_str = stress_pattern.findall(block)[0]
            virial_stress = (
                np.array(list(map(formatify, stress_str.split())))
                .reshape(
                    6,
                )
                .tolist()
            )
            virial_stress = [virial_stress[self.mtp_stress_order.index(n)] for n in self.vasp_stress_order]
            struct = Structure(lattice=lattice, species=species, coords=position, coords_are_cartesian=True)
            d["structure"] = struct.as_dict()
            d["outputs"]["energy"] = energy
            assert size == struct.num_sites
            d["num_atoms"] = size
            d["outputs"]["forces"] = forces
            d["outputs"]["virial_stress"] = virial_stress

            data_pool.append(d)
        _, df = convert_docs(docs=data_pool, include_stress=True)
        return data_pool, df

    def train(
        self,
        train_structures,
        train_energies,
        train_forces,
        train_stresses,
        unfitted_mtp="08g.mtp",
        max_dist=5,
        radial_basis_size=8,
        max_iter=1000,
        energy_weight=1,
        force_weight=1e-2,
        stress_weight=1e-3,
        init_params="same",
        scale_by_force=0,
        bfgs_conv_tol=1e-3,
        weighting="vibration",
    ):
        """
        Training data with moment tensor method.

        Args:
            train_structures ([Structure]): The list of Pymatgen Structure object.
                energies ([float]): The list of total energies of each structure
                in structures list.
            train_energies ([float]): List of total energies of each structure in
                structures list.
            train_forces ([np.array]): List of (m, 3) forces array of each structure
                with m atoms in structures list. m can be varied with each single
                structure case.
            train_stresses (list): List of (6, ) virial stresses of each structure
                in structures list.
            unfitted_mtp (str): Define the initial mtp file. Default to the mtp file
                stored in .params directory.
            max_dist (float): The actual radial cutoff.
            radial_basis_size (int): Relevant to number of radial basis function.
            max_iter (int): The number of maximum iteration.
            energy_weight (float): The weight of energy.
            force_weight (float): The weight of forces.
            stress_weight (float): The weight of stresses. Zero-weight can be assigned.
            init_params (str): How to initialize parameters if a potential was not
                pre-fitted. Choose from "same" and "random".
            scale_by_force (float): Default=0. If >0 then configurations near equilibrium
               (with roughtly force < scale_by_force) get more weight.
            bfgs_conv_tol (float): Stop training if error dropped by a factor smaller than this
                over 50 BFGS iterations.
            weighting (str): How to weight configuration with different sizes relative to each other.
                Choose from "vibrations", "molecules" and "structures".
        """
        if not which("mlp"):
            raise RuntimeError(
                "mlp has not been found.\n",
                "Please refer to https://mlip.skoltech.ru",
                "for further detail.",
            )
        train_structures, train_forces, train_stresses = check_structures_forces_stresses(
            train_structures, train_forces, train_stresses
        )
        train_pool = pool_from(train_structures, train_energies, train_forces, train_stresses)
        elements = sorted(set(itertools.chain(*[struct.species for struct in train_structures])))
        self.elements = [str(element) for element in elements]

        atoms_filename = "train.cfgs"

        with ScratchDir("."):
            atoms_filename = self.write_cfg(filename=atoms_filename, cfg_pool=train_pool)

            if not unfitted_mtp:
                raise RuntimeError("No specific parameter file provided.")
            MTP_file_path = os.path.join(module_dir, "params", unfitted_mtp)
            shutil.copyfile(MTP_file_path, os.path.join(os.getcwd(), unfitted_mtp))

            with open("min_dist", "w") as f:
                with subprocess.Popen(["mlp", "mindist", atoms_filename], stdout=f) as p:
                    p.communicate()[0]

            with open("min_dist") as f:
                lines = f.readlines()
            min_dist = float(lines[-1].split(":")[1])

            with open(unfitted_mtp) as f:
                template = f.read()

            s = template % (len(self.elements), min_dist, max_dist, radial_basis_size)
            with open(unfitted_mtp, "w") as f:
                f.write(s)

            save_fitted_mtp = ".".join([unfitted_mtp.split(".")[0] + "_fitted", unfitted_mtp.split(".")[1]])

            with subprocess.Popen(
                [
                    "mlp",
                    "train",
                    unfitted_mtp,
                    atoms_filename,
                    f"--max-iter={max_iter}",
                    f"--trained-pot-name={save_fitted_mtp}",
                    f"--curr-pot-name={unfitted_mtp}",
                    f"--energy-weight={energy_weight}",
                    f"--force-weight={force_weight}",
                    f"--stress-weight={stress_weight}",
                    f"--init-params={init_params}",
                    f"--scale-by-force={scale_by_force}",
                    f"--bfgs-conv-tol={bfgs_conv_tol}",
                    f"--weighting={weighting}",
                ],
                stdout=subprocess.PIPE,
            ) as p:
                stdout = p.communicate()[0]
                rc = p.returncode
            if rc != 0:
                error_msg = f"MLP exited with return code {rc}"
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)

            def load_config(filename):
                param = OrderedDict()
                with open(filename) as f:
                    lines = f.readlines()
                param["safe"] = [line.rstrip() for line in lines[:-2]]
                for line in lines[-2:]:
                    key = line.rstrip().split(" = ")[0]
                    value = json.loads(line.rstrip().split(" = ")[1].replace("{", "[").replace("}", "]"))
                    param[key] = value
                return param

            self.param = load_config(save_fitted_mtp)
        return rc

    def write_param(self, fitted_mtp="fitted.mtp", **kwargs):
        """
        Write fitted mtp parameter file to perform lammps calculation.

        Args:
            fitted_mtp (str): Filename to store xml formatted parameters.
        """
        if not self.param:
            raise RuntimeError("The parameters should be provided.")
        lines = [
            " = ".join([key, json.dumps(value).replace("[", "{").replace("]", "}")])
            if key != "safe"
            else "\n".join(value)
            for key, value in self.param.items()
        ]
        with open(fitted_mtp, "w") as f:
            f.write("\n".join(lines))
        ini_file = self.write_ini(load_from=fitted_mtp, Calculate_EFS=True, **kwargs)
        ff_settings = [self.pair_style.format(ini_file), self.pair_coeff]
        return ff_settings

    def evaluate(self, test_structures, test_energies, test_forces, test_stresses=None, **kwargs):
        """
        Evaluate energies, forces and stresses of structures with trained
        interatomic potentials.

        Args:
            test_structures ([Structure]): List of Pymatgen Structure Objects.
            test_energies ([float]): List of DFT-calculated total energies of
                each structure in structures list.
            test_forces ([np.array]): List of DFT-calculated (m, 3) forces of
                each structure with m atoms in structures list. m can be varied
                with each single structure case.
            test_stresses (list): List of DFT-calculated (6, ) viriral stresses
                of each structure in structures list.
            kwargs: Parameters of write_param method.
        """
        if not which("mlp"):
            raise RuntimeError(
                "mlp has not been found.\n",
                "Please refer to https://mlip.skoltech.ru ",
                "for further detail.",
            )
        fitted_mtp = "fitted.mtp"
        original_file = "original.cfgs"
        predict_file = "predict.cfgs"
        test_structures, test_forces, test_stresses = check_structures_forces_stresses(
            test_structures, test_forces, test_stresses
        )
        predict_pool = pool_from(test_structures, test_energies, test_forces, test_stresses)

        with ScratchDir("."):
            self.write_param(
                fitted_mtp=fitted_mtp,
                Abinitio=0,
                Driver=1,
                Write_cfgs=predict_file,
                Database_filename=original_file,
                **kwargs,
            )
            original_file = self.write_cfg(original_file, cfg_pool=predict_pool)
            _, df_orig = self.read_cfgs(original_file)

            if not hasattr(self, "version") or self.version == "mlip-2":
                cmd = ["mlp", "calc-efs", fitted_mtp, original_file, predict_file]
            elif self.version == "mlip-dev":
                cmd = ["mlp", "run", "mlip.ini", f"--filename={original_file}"]
            with subprocess.Popen(cmd, stdout=subprocess.PIPE) as p:
                stdout = p.communicate()[0]
                rc = p.returncode
            if rc != 0:
                error_msg = f"mlp exited with return code {rc}"
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)
            if not os.path.exists(predict_file):
                predict_file = "_".join([predict_file, "0"])
            _, df_predict = self.read_cfgs(predict_file)
        return df_orig, df_predict

    @staticmethod
    def from_config(filename, elements):
        """
        Initialize potentials with parameters file.

        Args:
            filename (str): The file storing parameters of potentials, filename should
                ends with ".mtp".
            elements (list): The list of elements.

        Returns:
            MTPotential
        """
        param = OrderedDict()
        with open(filename) as f:
            lines = f.readlines()
        param["safe"] = [line.rstrip() for line in lines[:-2]]
        for line in lines[-2:]:
            key = line.rstrip().split(" = ")[0]
            value = json.loads(line.rstrip().split(" = ")[1].replace("{", "[").replace("}", "]"))
            param[key] = value

        mtp = MTPotential(param=param)
        mtp.elements = elements

        return mtp



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""
Define abstract base classes.
"""
from ._data import BaseDataSource  # noqa
from ._describer import (  # noqa
    BaseDescriber,
    DummyDescriber,
    SequentialDescriber,
    describer_type,
)
from ._feature_batch import get_feature_batch
from ._mixin import TargetScalerMixin
from ._model import (  # noqa
    BaseModel,
    KerasModel,
    SKLModel,
    is_keras_model,
    is_sklearn_model,
)

__all__ = [
    "BaseDataSource",
    "BaseDescriber",
    "SequentialDescriber",
    "DummyDescriber",
    "BaseModel",
    "KerasModel",
    "SKLModel",
    "get_feature_batch",
    "describer_type",
    "is_keras_model",
    "is_sklearn_model",
    "TargetScalerMixin",
]



"""
Module implements the perturbation class for atomic and lattice relaxation.
"""
import os
from typing import List, Union

import numpy as np
from monty.serialization import loadfn
from pymatgen.core.operations import SymmOp
from pymatgen.core.sites import PeriodicSite, Site
from pymatgen.core.structure import Lattice, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.groups import SpaceGroup, in_array_list

module_dir = os.path.dirname(__file__)

wyckoff_nums = loadfn(os.path.join(module_dir, "symmetry_rules", "wyckoff_nums.json"))
wyckoff_nums = {int(k): v for k, v in wyckoff_nums.items()}

wyckoff_dims = loadfn(os.path.join(module_dir, "symmetry_rules", "wyckoff_dims.json"))
wyckoff_dims = {int(k): v for k, v in wyckoff_dims.items()}

standard_modes = loadfn(os.path.join(module_dir, "symmetry_rules", "standard_modes.json"))
standard_modes = {int(k): v for k, v in standard_modes.items()}

perturbation_modes = loadfn(os.path.join(module_dir, "symmetry_rules", "perturbation_modes.json"))
perturbation_modes = {int(k): v for k, v in perturbation_modes.items()}

small_addup = np.array([1e-4] * 3)


def perturbation_mapping(x, fixed_indices):
    """
    Perturbation mapping.

    Args:
        x:
        fixed_indices:

    Returns:

    """
    return np.array(
        [
            0 if i in fixed_indices else x[np.argwhere(np.arange(3)[~np.isin(range(3), fixed_indices)] == i)[0][0]]
            for i in range(3)
        ]
    )


class WyckoffPerturbation:
    """
    Perturbation class for determining the standard wyckoff position
    and generating corresponding equivalent fractional coordinates.
    """

    def __init__(
        self, int_symbol: int, wyckoff_symbol: str, symmetry_ops: List[SymmOp] = None, use_symmetry: bool = True
    ):
        """
        Args:
            int_symbol (int): International number of space group.
            wyckoff_symbol (str): Wyckoff symbol.
            symmetry_ops (list): Full set of symmetry operations as matrices.
                Use specific symmetry operations if initialized.
            use_symmetry (bool): Whether to use constraint of symmetry to reduce
                parameters space.
        """
        self._site = None
        self._fit_site = False
        self.int_symbol = int_symbol
        self.wyckoff_symbol = wyckoff_symbol
        self.use_symmetry = use_symmetry
        if self.use_symmetry:
            self.standard_mode = eval(standard_modes[int_symbol][wyckoff_symbol])
            self.dim = wyckoff_dims[int_symbol][wyckoff_symbol]
            self.multiplicity = dict(zip(*wyckoff_nums[int_symbol]))[wyckoff_symbol]
            self.perturbation_mode = eval(perturbation_modes[int_symbol][wyckoff_symbol])
            self.symmetry_ops = (
                SpaceGroup.from_int_number(int_symbol).symmetry_ops if not symmetry_ops else symmetry_ops
            )
        else:
            self.standard_mode = eval("lambda p: True")
            self.dim = 3
            self.multiplicity = 1
            self.perturbation_mode = eval("lambda x: x")
            self.symmetry_ops = SpaceGroup.from_int_number(1).symmetry_ops

    def get_orbit(self, p: Union[List, np.ndarray], tol: float = 1e-3) -> List[np.ndarray]:
        """
        Returns the orbit for a point.

        Args:
            p (list/numpy.array): Fractional coordinated point.
            tol (float): Tolerance for determining if sites are the same.
        """
        orbit: List[np.ndarray] = []
        for symm_op in self.symmetry_ops:
            pp = symm_op.operate(p)
            pp[(pp + np.ones(3) * tol) % 1.0 < tol] = 0.0
            pp = np.mod(np.round(pp, decimals=10), 1)
            if not in_array_list(orbit, pp, tol=tol):
                orbit.append(pp)

        return orbit

    def sanity_check(self, site: Union[Site, PeriodicSite], wyc_tol: float = 0.3 * 1e-3) -> None:
        """
        Check whether the perturbation mode exists.

        Args:
            site (PeriodicSite): PeriodicSite in Structure.
            wyc_tol (float): Tolerance for wyckoff symbol determined coordinates.
        """
        p = site.frac_coords
        orbits = self.get_orbit(p, wyc_tol)

        if len(orbits) != self.multiplicity:
            return

        for pp in orbits:
            if self.standard_mode(pp):
                self._site = site  # type: ignore
                self._fit_site = True
                break

    def standardize(self, p: Union[List, np.ndarray], tol: float = 1e-3) -> List[float]:
        """
        Get the standardized position of p.

        Args:
            p (list/numpy.array): Fractional coordinated point.
            tol (float): Tolerance for determining if sites are the same.
        """
        pp: List[float] = []
        orbits = self.get_orbit(p, tol)
        for pp in orbits:  # type: ignore
            if self.standard_mode(pp):
                break
        return pp

    @property
    def site(self):
        """
        Returns the site.
        """
        return self._site

    @property
    def fit_site(self):
        """
        Returns whether the site fits any standard wyckoff position.
        """
        return self._fit_site

    def __repr__(self):
        if self._site is not None:
            a, b, c = self._site.frac_coords
            return (
                f"{self.__class__.__name__}(spg_int_number={self.int_symbol}, wyckoff_symbol={self.wyckoff_symbol})"
                f" {self._site.species_string} [{a:.4f}, {b:.4f}, {c:.4f}]"
            )
        return f"{self.__class__.__name__}(spg_int_number={self.int_symbol}, wyckoff_symbol={self.wyckoff_symbol})"


def crystal_system(int_number: int) -> str:
    """
    Method for crystal system determination.

    Args:
        int_number (int): International number of space group.
    """
    if int_number <= 2:
        return "triclinic"
    if int_number <= 15:
        return "monoclinic"
    if int_number <= 74:
        return "orthorhombic"
    if int_number <= 142:
        return "tetragonal"
    if int_number <= 167 and int_number not in [
        143,
        144,
        145,
        147,
        149,
        150,
        151,
        152,
        153,
        154,
        156,
        157,
        158,
        159,
        162,
        163,
        164,
        165,
    ]:
        return "rhombohedral"
    if int_number <= 194:
        return "hexagonal"
    return "cubic"


class LatticePerturbation:
    """
    Perturbation class for determining the standard lattice.
    """

    def __init__(self, spg_int_symbol: int, use_symmetry: bool = True):
        """
        Args:
            spg_int_symbol (int): International number of space group.
            use_symmetry (bool): Whether to use constraint of symmetry to reduce
                parameters space.
        """
        self._lattice = None
        self._fit_lattice = False
        self.spg_int_symbol = spg_int_symbol
        self.use_symmetry = use_symmetry
        self.crys_system = crystal_system(spg_int_symbol)

    def sanity_check(self, lattice: Lattice, abc_tol: float = 1e-3, angle_tol: float = 3e-1) -> None:
        """
        Check whether the perturbation mode exists.

        Args:
            lattice (Lattice): Lattice in Structure.
            abc_tol (float): Tolerance for lattice lengths determined by crystal system.
            angle_tol (float): Tolerance for lattice angles determined by crystal system.
        """
        abc = list(lattice.abc)
        angles = list(lattice.angles)

        def check(param, ref, tolerance):
            return all(abs(i - j) < tolerance for i, j in zip(param, ref) if j is not None)

        if not self.use_symmetry:
            self.dims = (3, 3)
            self.perturbation_mode = lambda x: x
            self._lattice = lattice  # type: ignore
            self._abc = abc
            self._fit_lattice = True
            return

        if self.crys_system == "cubic":
            a = abc[0]
            if not (check(abc, [a, a, a], abc_tol) and check(angles, [90, 90, 90], angle_tol)):
                self._fit_lattice = False
                return
            self.dims = (1, 0)
            self.perturbation_mode = lambda x: np.concatenate((np.repeat(x, repeats=3), np.zeros(3)))
            self._lattice = lattice  # type: ignore
            self._abc = [abc[0]]  # type: ignore
            self._fit_lattice = True
            return
        if self.crys_system == "hexagonal":
            if not (
                np.any([(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)])
                and check(np.sort(angles), [90, 90, 120], angle_tol)
            ):
                self._fit_lattice = False
                return
            self.dims = (2, 0)
            indices = [int(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]
            self.perturbation_mode = lambda x: np.concatenate((x[indices], np.zeros(3)))
            self._lattice = lattice  # type: ignore
            self._abc = [
                abc[indices.index(0)],  # type: ignore
                abc[indices.index(1)],
            ]  # type: ignore
            self._fit_lattice = True
            return
        if self.crys_system == "rhombohedral":
            a = abc[0]
            alpha = angles[0]
            if check(abc, [a, a, a], abc_tol) and check(angles, [alpha, alpha, alpha], angle_tol):
                self.dims = (1, 1)
                self.perturbation_mode = lambda x: np.concatenate(
                    (np.repeat(x[0], repeats=3), np.repeat(x[1], repeats=3))
                )
                self._lattice = lattice  # type: ignore
                self._abc = [a]  # type: ignore
                self._fit_lattice = True
                return
            if np.any([(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]) and check(
                np.sort(angles), [90, 90, 120], angle_tol
            ):
                self.dims = (2, 0)
                indices = [int(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]
                self.perturbation_mode = lambda x: np.concatenate((x[indices], np.zeros(3)))
                self._lattice = lattice  # type: ignore
                self._abc = [
                    abc[indices.index(0)],  # type: ignore
                    abc[indices.index(1)],
                ]  # type: ignore
                self._fit_lattice = True
                return
            self._fit_lattice = False
            return
        if self.crys_system == "tetragonal":
            if not check(angles, [90, 90, 90], angle_tol):
                self._fit_lattice = False
                return
            if np.any([(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]):
                self.dims = (2, 0)
                indices = [int(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 2) for i in np.arange(3)]
                self.perturbation_mode = lambda x: np.concatenate((x[indices], np.zeros(3)))
                self._lattice = lattice  # type: ignore
                self._abc = [
                    abc[indices.index(0)],  # type: ignore
                    abc[indices.index(1)],
                ]  # type: ignore
                self._fit_lattice = True
                return
            if np.all([(sum(abs(np.array(abc) - abc[i]) < abc_tol) == 3) for i in np.arange(3)]):
                self.dims = (1, 0)
                self.perturbation_mode = lambda x: np.concatenate((np.repeat(x, repeats=3), np.zeros(3)))
                self._lattice = lattice  # type: ignore
                self._abc = [abc[0]]  # type: ignore
                self._fit_lattice = True
                return
            self._fit_lattice = False
            return
        if self.crys_system == "orthorhombic":
            if not check(angles, [90, 90, 90], angle_tol):
                self._fit_lattice = False
                return
            self.dims = (3, 0)
            self.perturbation_mode = lambda x: np.concatenate((x, np.zeros(3)))
            self._lattice = lattice  # type: ignore
            self._abc = list(abc)  # type: ignore
            self._fit_lattice = True
            return
        if self.crys_system == "monoclinic":
            if sum(abs(angles[i] - 90) < 1e-5 for i in np.arange(3)) == 2:
                self.dims = (3, 1)
                indices = [int(abs(angles[i] - 90) < angle_tol) for i in np.arange(3)]
                self.perturbation_mode = lambda x: np.concatenate(
                    (
                        np.pad(x, pad_width=(0, 5 - len(x)), mode="constant")[:3],
                        np.pad(x, pad_width=(0, 5 - len(x)), mode="constant")[3:][indices],
                    )
                )
                self._lattice = lattice  # type: ignore
                self._abc = list(abc)  # type: ignore
                self._fit_lattice = True
                return
            if check(angles, [90, 90, 90], angle_tol):
                self.dims = (3, 0)
                self.perturbation_mode = lambda x: np.concatenate((x, np.zeros(3)))
                self._lattice = lattice  # type: ignore
                self._abc = list(abc)  # type: ignore
                self._fit_lattice = True
                return
            self._fit_lattice = False
            return
        self.dims = (3, 3)
        self.perturbation_mode = lambda x: x
        self._lattice = lattice  # type: ignore
        self._abc = list(abc)  # type: ignore
        self._fit_lattice = True
        return

    @property
    def fit_lattice(self) -> bool:
        """
        Returns whether the lattice fits any crystal system.
        """
        return self._fit_lattice

    @property
    def lattice(self) -> Lattice:
        """
        Returns the lattice.
        """
        return self._lattice  # type: ignore

    @property
    def abc(self) -> List[float]:
        """
        Returns the lattice lengths.
        """
        return self._abc  # type: ignore

    def __repr__(self):
        if self._lattice is not None:
            return (
                f"{self.__class__.__name__}(spg_int_number={self.spg_int_symbol}, "
                f"crystal_system={self.crys_system})\n" + repr(self.lattice)
            )
        return f"{self.__class__.__name__}(spg_int_number={self.spg_int_symbol}, crystal_system={self.crys_system})\n"


def get_standardized_structure(structure: Structure) -> Structure:
    """
    Get standardized structure.

    Args:
        structure (Structure): Pymatgen Structure object.
    """
    species = [dict(site.species.as_dict()) for site in structure]
    sa = SpacegroupAnalyzer(structure)
    sd = sa.get_symmetry_dataset()
    std_lattice = sd["std_lattice"]
    mapping_to_primitive = sd["mapping_to_primitive"].tolist()
    mapping_to_primitive = np.array([mapping_to_primitive.index(i) for i in range(max(mapping_to_primitive) + 1)])
    std_species = np.array(species)[mapping_to_primitive][sd["std_mapping_to_primitive"]]
    std_frac_coords = sd["std_positions"]
    return Structure(lattice=std_lattice, species=std_species, coords=std_frac_coords).get_sorted_structure()



"""
Module implemets the target space.
"""
from typing import Callable, Dict, List, Union

import numpy as np
from numpy.random import RandomState
from pymatgen.core.periodic_table import _pt_data as pt_data

from maml.apps.bowsr.acquisition import lhs_sample
from maml.apps.bowsr.perturbation import WyckoffPerturbation
from maml.apps.bowsr.preprocessing import DummyScaler, StandardScaler


def _hashable(x):
    """
    Ensure that an point is hashable by a python dict.
    """
    return tuple(map(float, x))


class TargetSpace:
    """
    Holds the perturbations of coordinates/lattice (x_wyckoff/x_lattice)
    -- formation energy (Y). Allows for constant-time appends while
    ensuring no duplicates are added.
    """

    def __init__(
        self,
        target_func: Callable,
        wps: List[WyckoffPerturbation],
        abc_dim: int,
        angles_dim: int,
        relax_coords: bool,
        relax_lattice: bool,
        scaler: Union[StandardScaler, DummyScaler],
        random_state: RandomState,
    ):
        """
        Args:
            target_func: Target function to be maximized.
            wps (list): WyckoffPerturbations for derivation of symmetrically
                unique sites. Used to derive the coordinates of the sites.
            abc_dim (int): Number of independent variables in lattice constants.
            angles_dim (int): Number of independent variables in lattice angles.
            relax_coords (bool): Whether to relax symmetrical coordinates.
            relax_lattice (bool): Whether to relax lattice.
            scaler: Scaler used in scaling params.
            random_state (RandomState): Seeded rng for random number generator. None
                for an unseeded rng.
        """
        self.target_func = target_func
        self.random_state = random_state
        self.wps = wps
        self.abc_dim = abc_dim
        self.angles_dim = angles_dim
        self.relax_coords = relax_coords
        self.relax_lattice = relax_lattice
        self.scaler = scaler
        wyckoff_dims = [wp.dim for wp in wps]

        if not any([relax_coords, relax_lattice]):
            raise ValueError(
                "No degrees of freedom for relaxation are given. "
                "Please check the relax_coords and relax_lattice arguments."
            )

        if all([relax_coords, relax_lattice]):
            self.dim = sum(wyckoff_dims) + abc_dim + angles_dim
        elif relax_coords:
            self.dim = sum(wyckoff_dims)
        else:
            self.dim = abc_dim + angles_dim

        # Preallocated memory for x and y points.
        self._params = np.empty(shape=(0, self.dim))
        self._target = np.empty(shape=(0))

    def __len__(self):
        assert len(self._params) == len(self._target)
        return len(self._target)

    @property
    def params(self) -> np.ndarray:
        """
        Returns the parameters in target space.
        """
        return self._params

    @property
    def target(self) -> np.ndarray:
        """
        Returns the target (i.e., formation energy) in target space.
        """
        return self._target

    @property
    def bounds(self) -> np.ndarray:
        """
        Returns the search space of parameters.
        """
        return self._bounds

    def register(self, x, target) -> None:
        """
        Append a point and its target value to the known data.
        Args:
            x (ndarray): A single query point.
            target (float): Target value.
        """
        # if x in self:
        #     raise ValueError("Data point {} is not unique.".format(x))

        # Insert data into unique dictionary.
        x = self.scaler.transform(x)
        self._params = np.concatenate([self._params, x.reshape(1, -1)])
        self._target = np.concatenate([self._target, [target]])

    def probe(self, x) -> float:
        """
        Evaluate a single point x, to obtain the value y and then records
        them as observations.

        Args:
            x (ndarray): A single point.
        """
        target = self.target_func(x)
        self.register(x, target)
        return target

    def uniform_sample(self) -> np.ndarray:
        """
        Creates random points within the bounds of the space.
        """
        data = np.empty((1, self.dim))
        for col, (lower, upper) in enumerate(self.bounds):
            data.T[col] = np.round(self.random_state.uniform(lower, upper, size=1), decimals=3)
        return data.ravel()

    def lhs_sample(self, n_intervals: int) -> np.ndarray:
        """
        Latin hypercube sampling.

        Args:
            n_intervals (int): Number of intervals.
        """
        params = lhs_sample(n_intervals, self.bounds, self.random_state)
        return params

    def set_bounds(
        self, abc_bound: float = 1.2, angles_bound: float = 5, element_wise_wyckoff_bounds: Dict = None
    ) -> None:
        """
        Set the bound value of wyckoff perturbation and
        lattice perturbation/volume perturbation.
        """
        if not element_wise_wyckoff_bounds:
            element_wise_wyckoff_bounds = {}
        included_elements = list(element_wise_wyckoff_bounds.keys())
        element_wise_wyckoff_bounds = {
            frozenset({el: 1}.items()): bound for el, bound in element_wise_wyckoff_bounds.items()
        }
        element_wise_wyckoff_bounds.update({frozenset({el: 1}): 0.2 for el in pt_data if el not in included_elements})

        for wp in self.wps:
            frozen_key = frozenset(wp.site.species.as_dict().items())
            if frozen_key not in element_wise_wyckoff_bounds:
                element_wise_wyckoff_bounds.update({frozen_key: 0.2})

        wyckoff_bounds = np.concatenate(
            [
                np.tile([-1, 1], (wp.dim, 1))
                * element_wise_wyckoff_bounds[frozenset(wp.site.species.as_dict().items())]
                for wp in self.wps
            ]
        )
        abc_bounds = np.tile([-1, 1], (self.abc_dim, 1)) * abc_bound
        angles_bounds = np.tile([-1, 1], (self.angles_dim, 1)) * angles_bound

        if not any([self.relax_coords, self.relax_lattice]):
            raise ValueError(
                "No degrees of freedom for relaxation are given. "
                "Please check the relax_coords and relax_lattice arguments."
            )
        if all([self.relax_coords, self.relax_lattice]):
            self._bounds = np.concatenate((wyckoff_bounds, abc_bounds, angles_bounds))
        elif self.relax_coords:
            self._bounds = wyckoff_bounds
        else:
            self._bounds = np.concatenate((abc_bounds, angles_bounds))

    def set_empty(self) -> None:
        """
        Empty the param, target of the space.
        """
        self._params = np.empty(shape=(0, self.dim))
        self._target = np.empty(shape=(0))

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(relax_coords={self.relax_coords}, relax_lattice={self.relax_lattice},"
            f" dim={self.dim}, length={len(self)})"
        )



"""
This module implements more robust optmization
using the cvxpy package
"""
from typing import Dict, List, Optional, Union

import numpy as np
from monty.dev import requires
from scipy.linalg import lstsq

from maml.apps.symbolic._selectors import BaseSelector

try:
    import cvxpy as cp

    Expression = cp.expressions.expression.Expression
except ImportError:
    cp = None
    Expression = "Expression"


class BaseSelectorCP(BaseSelector):
    """
    Base selector using cvxpy (CP)
    """

    @requires(cp is not None, "cvxpy is not present.")
    def __init__(self, coef_thres: float = 1e-6, method: str = "ECOS"):
        """
        Base selector
        Args:
            coef_thres (float): threshold to discard certain coefficents
            method (str): solver for cvxpy problem. The default is ECOS
        """
        super().__init__(coef_thres=coef_thres, method=method)

    # pylint: disable=E1128
    def select(self, x: np.ndarray, y: np.ndarray, options: Optional[Dict] = None) -> Optional[np.ndarray]:
        """
        Select feature indices from x
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            options (dict): kwargs for cp.Problem.solve
        Returns: list of int indices
        """
        options = options or {}
        n, p = x.shape
        beta = cp.Variable(p)
        objective = cp.Minimize(self.construct_loss(x, y, beta))
        constraints = self.construct_constraints(x, y, beta)
        prob = cp.Problem(objective, constraints)
        prob.solve(solver=self.method, **options)
        self.coef_ = beta.value
        self.indices = np.where(np.abs(beta.value) > self.coef_thres)[0]
        self.coef_[np.where(np.abs(self.coef_) <= self.coef_thres)[0]] = 0.0  # type: ignore
        return self.indices

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: Optional[cp.Variable] = None
    ) -> Optional[List[Expression]]:  # type: ignore
        """
        Get constraints dictionary from data, e.g.,
        {"func": lambda beta: fun(x, y, beta), "type": "ineq"}
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta: (np.ndarray): target variable for optimization
        Returns: dict of constraints
        """
        return None

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: cp.Variable) -> Expression:  # type: ignore
        """
        Get loss function from data and tentative coefficients beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: loss value
        """
        raise NotImplementedError


class DantzigSelectorCP(BaseSelectorCP):
    """
    Equation 11 in
    https://orfe.princeton.edu/~jqfan/papers/06/SIS.pdf
    and reference in https://projecteuclid.org/download/pdfview_1/euclid.aos/1201012958
    """

    @requires(cp is not None, "cvxpy is not present.")
    def __init__(self, lambd, sigma=1.0, **kwargs):
        """
        Dantzig selector
        Args:
            lamb: tunable parameter
            sigma: standard deviation of the error
        """
        self.lambd = lambd
        self.sigma = sigma
        super().__init__(**kwargs)

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: cp.Variable) -> Expression:  # type: ignore
        """
        L1 loss
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (cp.Variable): dimension N vector for optimization
        Returns: loss expression
        """
        return cp.norm1(beta)

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: Optional[cp.Variable] = None
    ) -> Optional[List[Expression]]:  # type: ignore
        """
        Dantzig selector constraints
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (cp.Variable): dimension N vector for optimization
        Returns: List of constraints
        """
        return [cp.norm_inf(x.T @ (y - x @ beta)) <= self.lambd * self.sigma]


class PenalizedLeastSquaresCP(BaseSelectorCP):
    """
    Penalized least squares. In addition to minimizing the sum of squares loss,
    it adds an additional penalty to the coefficients
    """

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: cp.Variable) -> Expression:  # type: ignore
        """
        L1 loss
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (cp.Variable): dimension N vector for optimization
        Returns: loss expression
        """
        n = x.shape[0]
        se = 1.0 / (2 * n) * cp.sum_squares(y - x @ beta) + self.penalty(beta, x=x, y=y)
        return se

    def penalty(
        self, beta: cp.Variable, x: Optional[np.ndarray] = None, y: Optional[np.ndarray] = None
    ) -> Union[Expression, float]:  # type: ignore
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: penalty value
        """
        return 0.0


class LassoCP(PenalizedLeastSquaresCP):
    """
    Simple Lasso regression
    """

    @requires(cp is not None, "cvxpy not installed")
    def __init__(self, lambd, **kwargs):
        """
        Lasso regression with lambda * norm_1(beta) as penalty
        Args:
            lambd (float): weights for the penalty
            **kwargs:
        """
        self.lambd = lambd
        super().__init__(**kwargs)

    def penalty(
        self, beta: cp.Variable, x: Optional[np.ndarray] = None, y: Optional[np.ndarray] = None
    ) -> Union[Expression, float]:  # type: ignore
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: penalty value
        """
        beta_abs = cp.norm1(beta)
        return self.lambd * beta_abs


class AdaptiveLassoCP(PenalizedLeastSquaresCP):
    """
    Adaptive lasso regression using OLS coefficients
    as the root-n estimator coefficients
    """

    @requires(cp is not None, "cvxpy not installed")
    def __init__(self, lambd, gamma, **kwargs):
        """
        Adaptive lasso regression
        Args:
            lambd (float or list of floats):
            gamma (float): exponential for hat(beta)
            **kwargs:
        """
        self.lambd = lambd
        self.gamma = gamma
        self.w = 1
        super().__init__(**kwargs)

    def select(self, x: np.ndarray, y: np.ndarray, options: Optional[Dict] = None) -> Optional[np.ndarray]:
        """
        Select feature indices from x
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            options (dict): options in the cp.Problem.solve
        Returns: array int indices
        """
        self.w = self.get_w(x, y)
        return super().select(x, y, options)

    def get_w(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Get adaptive weights from data
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: coefficients array
        """
        beta_hat = lstsq(x, y)[0]
        w = 1.0 / np.abs(beta_hat) ** self.gamma
        return w

    def penalty(
        self, beta: cp.Variable, x: Optional[np.ndarray] = None, y: Optional[np.ndarray] = None
    ) -> Union[Expression, float]:  # type: ignore
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: penalty value
        """
        return cp.sum(self.lambd * cp.multiply(self.w, cp.abs(beta)))



"""
Describer for converting (structural) objects into models-readable
numeric vectors or tensors.
"""

from ._composition import ElementProperty, ElementStats  # noqa
from ._site import SmoothOverlapAtomicPosition  # noqa
from ._site import (  # noqa
    BispectrumCoefficients,
    BPSymmetryFunctions,
    MEGNetSite,
    SiteElementProperty,
)
from ._structure import CoulombMatrix  # noqa
from ._structure import SortedCoulombMatrix  # noqa
from ._structure import (
    CoulombEigenSpectrum,
    DistinctSiteProperty,
    MEGNetStructure,
    RandomizedCoulombMatrix,
)

__all__ = [
    "BispectrumCoefficients",
    "SmoothOverlapAtomicPosition",
    "BPSymmetryFunctions",
    "DistinctSiteProperty",
    "CoulombMatrix",
    "SortedCoulombMatrix",
    "RandomizedCoulombMatrix",
    "ElementProperty",
    "ElementStats",
    "MEGNetSite",
    "MEGNetStructure",
    "CoulombEigenSpectrum",
    "SiteElementProperty",
]



"""
DFT wrapper
"""
import os
import subprocess

from monty.os.path import which
from monty.serialization import loadfn
from monty.tempfile import ScratchDir
from pymatgen.core.structure import Structure
from pymatgen.entries.compatibility import MaterialsProjectCompatibility
from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.io.vasp.sets import MPStaticSet

from maml.apps.bowsr.model.base import EnergyModel

module_dir = os.path.dirname(__file__)
elements_filename = os.path.join(module_dir, "..", "regularization", "elements.json")
elements = loadfn(elements_filename)


class DFT(EnergyModel):
    """
    DFT static calculation wrapped as energy model.
    """

    def __init__(self, exe_path: str = None):
        """
        DFT wrapper
        Args:
            exe_path: VASP executable path
        """
        if not exe_path:
            if not which("vasp_std"):
                raise RuntimeError("Vasp executable can not be found.")
            self.vasp_exe = which("vasp_std")
        else:
            self.vasp_exe = exe_path

    def predict_energy(self, structure: Structure):
        """
        Predict energy from structure.
        Args:
            structure: (pymatgen Structure)

        Returns: float
        """
        el_amt_dict = structure.composition.get_el_amt_dict()

        static = MPStaticSet(structure)
        with ScratchDir("."):
            static.write_input(".")
            with subprocess.Popen([self.vasp_exe], stdout=subprocess.PIPE, stderr=subprocess.PIPE) as p_exe:
                stdout, stderr = p_exe.communicate()
                rc = p_exe.returncode
            if rc != 0:
                error_msg = f"vasp exited with return code {rc}"
                msg = stderr.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = [i for i, m in enumerate(msg) if m.startswith("ERROR")][0]
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += ", "
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)

            compat = MaterialsProjectCompatibility()
            vrun = Vasprun("vasprun.xml")
            entry = compat.process_entry(vrun.get_computed_entry())
            energy = (
                entry.energy - sum(elements[el]["energy_per_atom"] * amt for el, amt in el_amt_dict.items())
            ) / len(structure)

        return energy



"""
Base class that expose a predict_energy method
"""
from pymatgen.core.structure import Structure


class EnergyModel:
    """
    Base energy model class. For any model used in BOWSR, it has to have
    a predict_energy method that returns a float
    """

    def predict_energy(self, structure: Structure) -> float:
        """
        Predict the energy of a structure
        Args:
            structure: Pymatgen Structure object.

        Returns: (float) energy value.

        """
        raise NotImplementedError



"""
Symbolic learning
"""
try:
    import cvxpy as cp
except ImportError:
    cp = None

from ._feature_generator import FeatureGenerator, Operator  # noqa
from ._selectors import SCAD  # noqa
from ._selectors import L0BrutalForce  # noqa
from ._sis import ISIS, SIS  # noqa

if cp is None:
    from ._selectors import AdaptiveLasso, DantzigSelector, Lasso  # noqa  # noqa
else:
    # import from cvxpy alternatives
    from ._selectors_cvxpy import AdaptiveLassoCP as AdaptiveLasso  # noqa
    from ._selectors_cvxpy import DantzigSelectorCP as DantzigSelector  # noqa
    from ._selectors_cvxpy import LassoCP as Lasso  # noqa

__all__ = [
    "DantzigSelector",
    "AdaptiveLasso",
    "SCAD",
    "Lasso",
    "SIS",
    "ISIS",
    "L0BrutalForce",
    "FeatureGenerator",
    "Operator",
]



"""Signal processing utils"""
from math import ceil, floor
from typing import Callable, Tuple, Union

import numpy as np
from monty.dev import requires
from scipy import fft, signal

try:
    import tftb
except ImportError:
    tftb = None


def fft_magnitude(z: np.ndarray) -> np.ndarray:
    """
    Dicrete Fourier Transform the signal z and return
    the magnitude of the  coefficents
    Args:
        z (np.ndarray): 1D signal array
    Returns: 1D magnitude
    """
    return np.absolute(fft.fft(z))


def spectrogram(z: np.ndarray, return_time_freq: bool = False) -> Union[Tuple, np.ndarray]:
    """
    The spectrogram of the signal
    Args:
        z (np.ndarray): 1D signal array
        return_time_freq (bool): whether to return time and frequency
    Returns: 2D spectrogram
    """
    nx = len(z)
    nsc = floor(nx / 4.5)  # use matlab values
    nov = floor(nsc / 2)
    nfft = max(256, 2 ** ceil(np.log2(nsc)))
    freq, time, s = signal.spectrogram(z, window=signal.get_window("hann", nsc), noverlap=nov, nfft=nfft)
    if return_time_freq:
        return freq, time, s
    return s


def cwt(z: np.ndarray, widths: np.ndarray, wavelet: Union[str, Callable] = "morlet2", **kwargs) -> np.ndarray:
    """
    The scalogram of the signal
    Args:
        z (np.ndarray): 1D signal array
        widths (np.ndarray): wavelet widths
        wavelet (str): wavelet name
    Returns: 2D scalogram
    """
    if isinstance(wavelet, str):
        wavelet_func = getattr(signal, wavelet)
    else:
        wavelet_func = wavelet
    return np.absolute(signal.cwt(z, wavelet_func, widths=widths, **kwargs))


@requires(tftb is not None, "Requires installation of tftb package")
def wvd(z: np.ndarray, return_all: bool = False) -> Union[Tuple, np.ndarray]:
    """
    Wigner Ville Distribution calculator
    Args:
        z (np.ndarray): signal 1D
        return_all (bool): whether to return time and freq info, default
            only return the wvd information
    Returns: NxN wvd matrix
    """
    tfr = tftb.processing.WignerVilleDistribution(z)
    (
        res,
        f1,
        f2,
    ) = tfr.run()
    if return_all:
        return res, f1, f2
    return res


AVAILABLE_SP_METHODS = {"fft_magnitude": fft_magnitude, "spectrogram": spectrogram, "cwt": cwt, "wvd": wvd}


def get_sp_method(sp_method: Union[str, Callable]) -> Callable:  # type: ignore
    """
    Providing a signal processing method name return the callable
    Args:
        sp_method (str): name of the sp function
    Returns: callable for signal processing
    """
    if isinstance(sp_method, str):
        try:
            return AVAILABLE_SP_METHODS[sp_method]
        except KeyError:
            raise KeyError(f"{sp_method} is not in available methods: " f"{AVAILABLE_SP_METHODS.keys()}")
    else:
        return sp_method



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This module provides SNAP interatomic potential class."""

import re
from logging import getLogger

import numpy as np
from monty.io import zopen
from sklearn.linear_model import LinearRegression

from maml.base import SKLModel
from maml.describers import BispectrumCoefficients
from maml.utils import (
    check_structures_forces_stresses,
    convert_docs,
    pool_from,
    stress_format_change,
)

from ._lammps import LammpsPotential

logger = getLogger(__name__)


class SNAPotential(LammpsPotential):
    """
    This class implements Spectral Neighbor Analysis Potential.
    """

    pair_style = "pair_style        snap"
    pair_coeff = "pair_coeff        * * {coeff_file} {param_file} {elements}"

    def __init__(self, model, name=None):
        """
        Initialize the SNAPotential Potential with atomic describers
        and models, which are used to generate the Bispectrum coefficients
        features for structures and to train the parameters.

        Args:
            model (LinearModel): Model to perform supervised learning with
                atomic descriptos as features and properties as targets.
            name (str): Name of force field.
        """

        logger.warning(
            "Triclinic structures will be rotated to lammps format. "
            "Please be sure to rotate forces and stresses to get the "
            "correct mapping for ensuring correct mapping. You may "
            "use`maml.utils.check_structures_forces_stresses` to do the "
            "correct rotations."
        )

        self.name = name if name else "SNAPotential"
        self.model = model
        self.elements = self.model.describer.elements

    def train(
        self,
        train_structures,
        train_energies,
        train_forces,
        train_stresses=None,
        include_stress=False,
        stress_format="VASP",
        **kwargs,
    ):
        """
        Training data with models.

        Args:
            train_structures ([Structure]): The list of Pymatgen Structure object.
                energies ([float]): The list of total energies of each structure
                in structures list.
            train_energies ([float]): List of total energies of each structure in
                structures list.
            train_forces ([np.array]): List of (m, 3) forces array of each
                structure with m atoms in structures list. m can be varied with
                each single structure case.
            train_stresses (list): List of (6, ) virial stresses of each
                structure in structures list.
            include_stress (bool): Whether to include stress components.
            stress_format (string): stress format, default to VASP
        """
        train_structures, train_forces, train_stresses = check_structures_forces_stresses(
            train_structures, train_forces, train_stresses, stress_format=stress_format
        )
        if include_stress:
            train_stresses = [
                stress_format_change(i, from_format=stress_format, to_format="SNAP") for i in train_stresses
            ]

        train_pool = pool_from(train_structures, train_energies, train_forces, train_stresses)
        _, df = convert_docs(train_pool, include_stress=include_stress)
        ytrain = df["y_orig"] / df["n"]
        xtrain = self.model.describer.transform(train_structures)
        self.model.fit(features=xtrain, targets=ytrain, **kwargs)

    def evaluate(
        self,
        test_structures,
        test_energies,
        test_forces,
        test_stresses=None,
        include_stress=False,
        stress_format="VASP",
    ):
        """
        Evaluate energies, forces and stresses of structures with trained
        machinea learning potentials.

        Args:
            test_structures ([Structure]): List of Pymatgen Structure Objects.
            test_energies ([float]): List of DFT-calculated total energies of
                each structure in structures list.
            test_forces ([np.array]): List of DFT-calculated (m, 3) forces of
                each structure with m atoms in structures list. m can be varied
                with each single structure case.
            test_stresses (list): List of DFT-calculated (6, ) viriral stresses
                of each structure in structures list.
            include_stress (bool): Whether to include stress components.
            stress_format (str): stress format, default to "VASP"
        """
        test_structures, test_forces, test_stresses = check_structures_forces_stresses(
            test_structures, test_forces, test_stresses
        )
        if include_stress:
            test_stresses = [
                stress_format_change(i, from_format=stress_format, to_format="SNAP") for i in test_stresses
            ]

        predict_pool = pool_from(test_structures, test_energies, test_forces, test_stresses)
        _, df_orig = convert_docs(predict_pool, include_stress=include_stress)
        _, df_predict = convert_docs(pool_from(test_structures), include_stress=include_stress)
        outputs = self.model.predict_objs(objs=test_structures)
        df_predict["y_orig"] = df_predict["n"] * outputs

        return df_orig, df_predict

    def write_param(self):
        """
        Write parameter and coefficient file to perform lammps calculation.
        """
        param_file = f"{self.name}.snapparam"
        coeff_file = f"{self.name}.snapcoeff"
        model = self.model
        describer = self.model.describer
        profile = describer.element_profile
        ne = len(self.elements)
        nbc = len(describer.subscripts)
        if describer.quadratic:
            nbc += int((1 + nbc) * nbc / 2)

        coeff_lines = []
        coeff_lines.append(f"{ne} {nbc + 1}")
        for element, coeff in zip(self.elements, np.split(model.model.coef_, ne)):
            coeff_lines.append(f"{element} {profile[element]['r']} {profile[element]['w']}")
            coeff_lines.extend([str(c) for c in coeff])
        with open(coeff_file, "w") as f:
            f.write("\n".join(coeff_lines))

        param_lines = []
        keys = ["rcutfac", "twojmax"]
        param_lines.extend([f"{k} {getattr(describer, k)}" for k in keys])
        param_lines.extend(["rfac0 0.99363", "rmin0 0"])
        param_lines.append(f"quadraticflag {int(describer.quadratic)}")
        param_lines.append("bzeroflag 0")
        with open(param_file, "w") as f:
            f.write("\n".join(param_lines))

        pair_style = self.pair_style
        pair_coeff = self.pair_coeff.format(
            elements=" ".join(self.elements), coeff_file=coeff_file, param_file=param_file
        )
        ff_settings = [pair_style, pair_coeff]
        return ff_settings

    @staticmethod
    def from_config(param_file, coeff_file, **kwargs):
        """
        Initialize potentials with parameters file and coefficient file.

        Args:
            param_file (str): The file storing the configuration of potentials.
            coeff_file (str): The file storing the coefficients of potentials.

        Return:
            SNAPotential.
        """
        with open(coeff_file) as f:
            coeff_lines = f.readlines()
        coeff_lines = [line for line in coeff_lines if not line.startswith("#")]
        element_profile = {}
        ne, nbc = coeff_lines[0].split()
        ne, nbc = int(ne), int(nbc)
        for n in range(ne):
            specie, r, w = coeff_lines[1 + n * (nbc + 1)].split()
            r, w = float(r), float(w)
            element_profile[specie] = {"r": r, "w": w}

        rcut_pattern = re.compile(r"rcutfac (.*?)\n", re.S)
        twojmax_pattern = re.compile(r"twojmax (\d*)\n", re.S)
        quadratic_pattern = re.compile(r"quadraticflag (.*?)(?=\n|$)", re.S)

        with zopen(param_file, "rt") as f:
            param_lines = f.read()

        rcut = float(rcut_pattern.findall(param_lines)[-1])
        twojmax = int(twojmax_pattern.findall(param_lines)[-1])
        if quadratic_pattern.findall(param_lines):
            quadratic = bool(int(quadratic_pattern.findall(param_lines)[-1]))
        else:
            quadratic = False

        describer = BispectrumCoefficients(
            rcutfac=rcut, twojmax=twojmax, element_profile=element_profile, quadratic=quadratic, pot_fit=True
        )
        model = SKLModel(model=LinearRegression(), describer=describer, **kwargs)
        coef = np.array(
            np.concatenate([coeff_lines[(2 + nbc * n + n) : (2 + nbc * (n + 1) + n)] for n in range(ne)]),
            dtype=np.float64,
        )
        model.model.coef_ = coef
        model.model.intercept_ = 0
        snap = SNAPotential(model=model)
        return snap



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This package contains PotentialMixin classes representing Interatomic Potentials."""

from ._base import Potential, PotentialMixin  # noqa
from ._gap import GAPotential  # noqa
from ._lammps import DefectFormation  # noqa
from ._lammps import ElasticConstant  # noqa
from ._lammps import EnergyForceStress  # noqa
from ._lammps import LatticeConstant  # noqa
from ._lammps import LMPRelaxationCalculator  # noqa
from ._lammps import LMPStaticCalculator  # noqa
from ._lammps import NudgedElasticBand  # noqa
from ._lammps import SpectralNeighborAnalysis  # noqa
from ._lammps import SurfaceEnergy  # noqa
from ._lammps import get_default_lmp_exe  # noqa; noqa
from ._mtp import MTPotential  # noqa
from ._nnp import NNPotential  # noqa
from ._snap import SNAPotential  # noqa

__all__ = [
    "Potential",
    "GAPotential",
    "MTPotential",
    "SNAPotential",
    "NNPotential",
    "LMPStaticCalculator",
    "EnergyForceStress",
    "SpectralNeighborAnalysis",
    "ElasticConstant",
    "LatticeConstant",
    "NudgedElasticBand",
    "DefectFormation",
    "get_default_lmp_exe",
]



"""
Utilities package.
"""
from ._data_conversion import convert_docs, pool_from, to_array  # noqa
from ._data_split import DataSplitter, ShuffleSplitter  # noqa
from ._dummy import feature_dim_from_test_system, get_describer_dummy_obj  # noqa
from ._inspect import get_full_args  # noqa
from ._jit import njit  # noqa
from ._lammps import stress_format_change  # noqa
from ._lammps import stress_list_to_matrix  # noqa
from ._lammps import write_data_from_structure  # noqa
from ._lammps import (  # noqa
    check_structures_forces_stresses,
    get_lammps_lattice_and_rotation,
    stress_matrix_to_list,
)
from ._material import to_composition  # noqa
from ._preprocessing import DummyScaler, Scaler, StandardScaler  # noqa
from ._signal_processing import (  # noqa
    cwt,
    fft_magnitude,
    get_sp_method,
    spectrogram,
    wvd,
)
from ._stats import (  # noqa
    STATS_KWARGS,
    Stats,
    get_full_stats_and_funcs,
    stats_list_conversion,
)
from ._tempfile import MultiScratchDir  # noqa
from ._value_profile import ConstantValue, LinearProfile, ValueProfile  # noqa

__all__ = [
    "pool_from",
    "convert_docs",
    "to_array",
    "ValueProfile",
    "ConstantValue",
    "LinearProfile",
    "njit",
    "MultiScratchDir",
    "DataSplitter",
    "ShuffleSplitter",
    "write_data_from_structure",
    "check_structures_forces_stresses",
    "stress_format_change",
    "stress_list_to_matrix",
    "stress_matrix_to_list",
    "get_lammps_lattice_and_rotation",
    "Stats",
    "STATS_KWARGS",
    "stats_list_conversion",
    "get_full_stats_and_funcs",
    "spectrogram",
    "cwt",
    "wvd",
    "fft_magnitude",
    "get_sp_method",
    "to_composition",
    "get_full_args",
    "get_describer_dummy_obj",
    "feature_dim_from_test_system",
    "StandardScaler",
    "Scaler",
    "DummyScaler",
]



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

from setuptools import find_packages, setup

long_desc = """
maml (MAterials Machine Learning) is a Python package that aims to provide useful high-level interfaces that make ML
for materials science as easy as possible.

The goal of maml is not to duplicate functionality already available in other packages. maml relies on well-established
packages such as scikit-learn and tensorflow for implementations of ML algorithms, as well as other materials science
packages such as `pymatgen <http://pymatgen.org>`_ and `matminer <http://hackingmaterials.lbl.gov/matminer/>`_ for
crystal/molecule manipulation and feature generation.

Official documentation at http://maml.ai/

Features
--------

1. Convert materials (crystals and molecules) into features. In addition to common compositional, site and structural
   features, we provide the following fine-grain local environment features.

 a) Bispectrum coefficients
 b) Behler Parrinello symmetry functions
 c) Smooth Overlap of Atom Position (SOAP)
 d) Graph network features (composition, site and structure)

2. Use ML to learn relationship between features and targets. Currently, the `maml` supports `sklearn` and `keras`
   models.

3. Applications:

 a) `pes` for modelling the potential energy surface, constructing surrogate models for property prediction.

  i) Neural Network Potential (NNP)
  ii) Gaussian approximation potential (GAP) with SOAP features
  iii) Spectral neighbor analysis potential (SNAP)
  iv) Moment Tensor Potential (MTP)

 b) `rfxas` for random forest models in predicting atomic local environments from X-ray absorption spectroscopy.

Installation
------------

Pip install via PyPI::

    pip install maml

To run the potential energy surface (pes), lammps installation is required you can install from source or from `conda`::

    conda install -c conda-forge/label/cf202003 lammps

The SNAP potential comes with this lammps installation. The GAP package for GAP and MLIP package for MTP are needed to run the corresponding potentials. For fitting NNP potential, the `n2p2` package is needed.

Install all the libraries from requirement.txt file::

    pip install -r requirements.txt

For all the requirements above::

    pip install -r requirements-ci.txt
    pip install -r requirements-optional.txt
    pip install -r requirements-dl.txt
    pip install -r requirements.txt

Citing
------
::

    @misc{maml,
        author = {Chen, Chi and Zuo, Yunxing and Ye, Weike and Ong, Shyue Ping},
        title = {{Maml - materials machine learning package}},
        year = {2020},
        publisher = {GitHub},
        journal = {GitHub repository},
        howpublished = {https://github.com/materialsvirtuallab/maml},
    }

For the ML-IAP package (`maml.pes`), please cite::

    Zuo, Y.; Chen, C.; Li, X.; Deng, Z.; Chen, Y.; Behler, J.; Csányi, G.; Shapeev, A. V.; Thompson, A. P.;
    Wood, M. A.; Ong, S. P. Performance and Cost Assessment of Machine Learning Interatomic Potentials.
    J. Phys. Chem. A 2020, 124 (4), 731–745. https://doi.org/10.1021/acs.jpca.9b08723.
"""

setup(
    name="maml",
    packages=find_packages(),
    version="2022.8.12",
    install_requires=["numpy", "scipy", "monty", "scikit-learn", "pandas", "pymatgen", "tqdm"],
    extras_requires={
        "maml.apps.symbolic._selectors_cvxpy": ["cvxpy"],
        "tensorflow": ["tensorflow>=2"],
        "tensorflow with gpu": ["tensorflow-gpu>=2"],
    },
    namespace_packages=["maml.apps"],
    author="Materials Virtual Lab",
    author_email="ongsp@eng.ucsd.edu",
    maintainer="Shyue Ping Ong",
    maintainer_email="ongsp@eng.ucsd.edu",
    url="http://maml.ai/",
    license="BSD",
    description="maml is a machine learning library for materials science.",
    long_description=long_desc,
    keywords=["materials", "science", "deep", "learning"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    include_package_data=True,
    package_data={
        "maml": [
            "describers/data/*.json",
            "describers/data/megnet_models/*.json" "describers/data/megnet_mdoels/*.hdf5",
        ]
    },
)



"""
megnet model wrapper implementation
"""
import os
import warnings

import numpy as np
from monty.dev import requires
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Structure

from maml.apps.bowsr.model.base import EnergyModel

try:
    import megnet
    from megnet.data.crystal import CrystalGraph
    from megnet.data.graph import GaussianDistance
    from megnet.models import MEGNetModel
except Exception:
    megnet = None
    MEGNetModel = None
    CrystalGraph = None
    GaussianDistance = None
    warnings.warn("megnet module should be installed to use this model type!")

module_dir = os.path.dirname(__file__)
model_filename = os.path.join(module_dir, "model_files", "megnet", "formation_energy.hdf5")


@requires(megnet is not None, "megnet package needs to be installed to use " "this module")
class MEGNet(EnergyModel):
    """
    MEGNetModel wrapper.
    """

    def __init__(self, model=None, reconstruct=False, **kwargs):
        """

        Args:
            model:  MEGNet energy model
            reconstruct: Whether to reconstruct the model (used in
                disordered model)
            **kwargs:
        """
        model = model or MEGNetModel.from_file(model_filename)
        gaussian_cutoff = kwargs.get("gaussian_cutoff", 6)
        radius_cutoff = kwargs.get("radius_cutoff", 5)
        npass = kwargs.get("npass", 2)
        self.reconstruct = reconstruct
        weights = model.get_weights()
        self.embedding = weights[0]
        if reconstruct:
            cg = CrystalGraph(
                bond_converter=GaussianDistance(np.linspace(0, gaussian_cutoff, 100), 0.5), cutoff=radius_cutoff
            )
            model_new = MEGNetModel(100, 2, 16, npass=npass, graph_converter=cg)
            model_new.set_weights(weights[1:])
            self.model = model_new
        else:
            self.model = model

    def predict_energy(self, structure: Structure) -> float:
        """
        Predict energy from structure
        Args:
            structure: (pymatgen Structure)

        Returns: float
        """
        if not structure.is_ordered and not self.reconstruct:
            raise ValueError("To predict properties of disordered structure, " "please set reconstruct=True")

        if self.reconstruct:
            species = [dict(site.species.as_dict()) for site in structure]
            structure_copy = structure.copy()
            structure_copy[:] = ""  # dummy variable
            graph = self.model.graph_converter.convert(structure_copy)

            atom = []
            for d in species:
                f = np.zeros(16)
                for el, amt in d.items():
                    f += self.embedding[Element(el).number] * amt
                atom.append(f)
            graph["atom"] = atom

            energy = self.model.predict_graph(graph)[0]
        else:
            energy = self.model.predict_structure(structure)[0]

        return energy
